import copy
import pickle
import threading
import time
import numpy as np
import zmq
from typing import Iterable
import blosc2
from typing import List
from npy_append_array import NpyAppendArray
from logging import getLogger
from datetime import datetime
import os
from hashlib import md5
from mkidgen3.server.misc import zpipe
from mkidgen3.util import format_bytes
from mkidgen3.funcs import convert_adc_raw_to_mv, raw_iq_to_unit, raw_phase_to_radian
from ..system_parameters import SYSTEM_BANDWIDTH
from mkidgen3.rfsocmemory import memfree_mib
from .feedline_config import FeedlineConfig, WaveformConfig, IFConfig
from ..mkidpynq import PHOTON_DTYPE
from ..system_parameters import N_CHANNELS, N_IQ_GROUPS, N_PHASE_GROUPS, N_POSTAGE_CHANNELS, \
    PHOTON_POSTAGE_WINDOW_LENGTH, MAXIMUM_DESIGN_COUNTRATE_PER_S
from functools import cached_property
from mkidgen3.mkidpynq import unpack_photons

from typing import Tuple
PHOTON_DTYPE_PACKED = np.uint64


class FRSClient:
    def __init__(self, url: str, command_port: int = 10000, data_port: int = 10001, status_port: int = 10002):
        """
        Args:
            url: ip address of rfsoc for ex: rfsoc4x2b.physics.ucsb.edu
            command_port: port for feedline server commands
            data_port: port for capture data server
            status_port: port for status server
        """
        self.url = url
        self.command_port = command_port
        self.data_port = data_port
        self.status_port = status_port

    @property
    def command_url(self):
        return f'tcp://{self.url}:{self.command_port}'

    @property
    def data_url(self):
        return f'tcp://{self.url}:{self.data_port}'

    @property
    def status_url(self):
        return f'tcp://{self.url}:{self.status_port}'

    def __str__(self):
        return f'FRS@{self.url}:{self.command_port}'

    def bequiet(self):
        ctx = zmq.Context().instance()
        with ctx.socket(zmq.REQ) as s:
            s.connect(self.command_url)
            s.send_pyobj(('bequiet', {}))
            return s.recv_pyobj()

    def order(self, *args):
        ctx = zmq.Context().instance()
        with ctx.socket(zmq.REQ) as s:
            s.connect(self.command_url)
            s.send_pyobj(args)
            return s.recv_pyobj()


class CaptureRequest:
    """
    A Capture Request object is generated by the client and submitted to the feedline readout server
    """
    STATUS_ENDPOINT = 'inproc://cap_stat.xsub'
    DATA_ENDPOINT = 'inproc://cap_data.xsub'
    FINAL_STATUSES = ('finished', 'aborted', 'failed')
    SUPPORTED_TAPS = ('postage', 'photon', 'adc', 'iq', 'filtphase', 'ddciq')

    @staticmethod
    def validate_channels(tap: str, chan: list):
        """
        Verify that the channel selection is compatible with the driver.

        Note that this is not done dynamically, no driver discovery is performed

        See mkidgen3.drivers.FilterPhase, .drivers.FilterIQ, and .trigger.PhotonPostageFilter
        for details about groups and monitor channels.

        Args:
            chan: A list of channels
            tap: A capture location

        Returns: True|False
        """
        tap = tap.lower()
        if 'iq' in tap:
            l = N_CHANNELS
            m = N_CHANNELS
        elif 'phase' in tap:
            l = N_CHANNELS
            m = N_CHANNELS
        elif tap == 'postage':
            l = N_POSTAGE_CHANNELS
            m = N_CHANNELS
        elif chan:
            l = 0  # no channels for photon or adc capture
        else:
            return False
        try:
            assert len(chan) <= l
            for c in chan:
                assert 0 <= c < m
        except AssertionError:
            return False
        return True

    def __init__(self, n: int, tap: str, feedline_config: FeedlineConfig,
                 feedline_server: FRSClient, channels: Iterable | int | None = None, file: str = None,
                 fail_saturation_frac: None | float | int = None, numpy_metric: None | str = None,
                 _compression_override: bool = None):
        """

        Args:
            n: the number of samples to capture, for photons n is the buffer rotation interval request in ms
            tap: the location to capture: 'adc'|'iq'|'phase'|'postage'|'photon'
            feedline_config: the FL configuration required by the capture
            feedline_server: the FRS to capture from
            channels: the phase, IQ, or postage resonator channels to capture
            file: an optional file to save the result to
            fail_saturation_frac: Take a pre-capture of ADC data and abort if the maximum I or Q magnitude is greater
                                  than the fraction*ADC_MAX_INT. Set to None, 0, or > 1 to disable.
            numpy_metric: a numpy function to apply across samples on a channel by channel basis. Must have `out` and
                          `axis` as input parameters. Math is done to I and Q respectively (complex math not supported).
                          For example: 'mean' will return a mean of the values for each channel for IQ and phase data.
            _compression_override: skips blosc compression on the server if True. False or None (default) results in
                                   data being compressed before sending.
        """

        tap = tap.lower()
        assert tap in CaptureRequest.SUPPORTED_TAPS
        self.fail_saturation_frac = float(fail_saturation_frac) if fail_saturation_frac else None
        self.nsamp = n  # n is treated as the buffer time in ms for photons, and has limits enforced by the driver
        self._last_status = None
        if channels is not None:
            try:
                channels = tuple(sorted(set(channels)))
            except TypeError:
                channels = (channels,)
            if not CaptureRequest.validate_channels(tap, channels):
                raise ValueError('Invalid channels specification')
        self.channels = channels
        self.tap = tap  # maybe add some error handling here
        self.feedline_config = feedline_config
        self.server = feedline_server
        self._status_socket = None
        self._data_socket = None
        self._established = False
        self._compression_override = _compression_override
        if numpy_metric:
            try:
                self.numpy_metric = str(numpy_metric)
                getattr(np, self.numpy_metric)
            except AttributeError:
                raise ValueError('Not a good metric')
        else:
            self.numpy_metric = None
        self.data_endpoint = file or type(self).DATA_ENDPOINT

    def __hash__(self):
        return int(md5(str((hash(self.feedline_config), self.tap, self.data_endpoint,
                            self.nsamp, self.server.command_url)).encode()).hexdigest(), 16)

    # def __enter__(self):
    #     self.establish()
    #
    # def __exit__(self, exc_type, exc_val, exc_tb):
    #     if isinstance(exc_type, zmq.error.ContextTerminated):
    #         self._data_socket=None
    #         self._status_socket=None
    #     else:
    #         self.destablish()

    def __del__(self):
        try:
            self.destablish()
        except zmq.error.ContextTerminated:
            pass

    def __str__(self):
        return f'CapReq {self.server}:{self.tap} {str(hash(self))}'

    def submit(self, ctx: zmq.Context = None):
        getLogger(__name__).info(f'Submitting {self}..')
        ctx = ctx or zmq.Context().instance()
        with ctx.socket(zmq.REQ) as s:
            s.connect(self.server.command_url)
            s.send_pyobj(('capture', self))
            getLogger(__name__).info(f'...submitted...')
            resp = s.recv_pyobj()
            getLogger(__name__).info(f'response: {resp}')
            return resp

    @property
    def type(self):
        return 'engineering' if self.tap not in ('photon', 'postage') else self.tap

    @property
    def id(self):
        return str(hash(self)).encode()

    def establish(self, context: zmq.Context = None):
        """
        Open up the outbound sockets for status and data. If data is going to a file, verify it ca be opened for writing and close it
        """
        if self._status_socket is not None or self._data_socket is not None:
            getLogger(__name__).warning(f'{self} already established')
            return
        context = context or zmq.Context.instance()
        self._status_socket = context.socket(zmq.PUB)
        self._status_socket.connect(self.STATUS_ENDPOINT)
        if self.data_endpoint.startswith('file://'):
            f = self.data_endpoint.lstrip('file://')
            if not os.access('/' + os.path.dirname(f), os.W_OK):
                raise IOError(f'Unable to write to {f}')
            return
        self._data_socket = context.socket(zmq.PUB)
        self._data_socket.connect(self.data_endpoint)
        self._send_status('established')

    def destablish(self):
        if self._status_socket is not None:
            log = True
            getLogger(__name__).debug(f'de-establishing status...')
        try:
            # NB the context or socket's linger setting determines how long any pending messages have to get sent off
            self._status_socket.close()
            self._status_socket = None
        except AttributeError:
            pass
        if self._data_socket is not None:
            log = True
            getLogger(__name__).debug(f'de-establishing data...')
        try:
            self._data_socket.close()
            self._data_socket = None
        except AttributeError:
            pass
        if log:
            getLogger(__name__).debug(f'...de-established.')

    def fail(self, message, raise_exception=False):
        if self.completed:
            return
        try:
            if self.data_endpoint.startswith('file://'):
                pass
            else:
                self._data_socket.send_multipart([self.id, b''])
            self._send_status('failed', message)
            self.destablish()
        except zmq.ZMQError as ez:
            getLogger(__name__).warning(f'Failed to send fail message {self} due to {ez}')
            if raise_exception:
                raise ez

    def finish(self, raise_exception=True):
        if self.completed:
            return
        try:
            if self.data_endpoint.startswith('file://'):
                pass
            else:
                self._data_socket.send_multipart([self.id, b''])
            self._send_status('finished', f'Finished at UTC {datetime.utcnow()}')
            self.destablish()
        except zmq.ZMQError as ez:
            getLogger(__name__).warning(f'Failed to send finished message {self} due to {ez}')
            if raise_exception:
                raise ez

    def abort(self, message, raise_exception=False):
        if self.completed:
            return
        try:
            self._data_socket.send_multipart([self.id, b''])
            self._send_status('aborted', message)
            self.destablish()
        except zmq.ZMQError as ez:
            getLogger(__name__).warning(f'Failed to send abort message {self} due to {ez}')
            if raise_exception:
                raise ez

    def send_data(self, data, status='', copy=False, compress=True):
        """
        Send a (chunk) of data out. By default, data is published as a blosc2 compressed array with the
        capture id as the subscription key. CaptureRequests created with a file destination  will be saved as
        uncompressed npy data, appended on successive calls, no data will be published.

        Data is shipped in the (compressed) format it is given.

        Do not call this function from multiple threads.
        Call .establish() in the calling thread before calling this function or after calling destablish()

        This function is intended to be zero-copy, is as of 10/27/23 and should be maintained as such.

        Args:
            data: a array of data to send out
            status: (optional) a status update message to send out
            copy: copy the array first, you probably don't want this. copying the array then freeing the
            underlying buffer causes segfaults with the pycharm debugger. Be aware that the property
            zmq.COPY_THRESHOLD will case messages to be copied even if copy is False if the message size
            is less than the threshold.
            compress: compress the data with blosc2, ignored for file destinations.

        Returns: None | zmq.MessageTracker None is returned if copying or the destination is a file. A
        segfault may occur if the data is deallocated prior to the messagetracker being done.
        """
        if not self._status_socket:
            raise RuntimeError('Status socket is not established.')
        if self.data_endpoint.startswith('file://'):
            with NpyAppendArray('/' + self.data_endpoint.lstrip('file://'), delete_if_exists=False) as x:
                x.append(data)
            self._send_status('capturing', status)
            return
        elif not self._data_socket:
            raise RuntimeError('Data socket is not established.')

        #        getLogger(__name__).debug(f'MiB Free: {memfree_mib()}')
        if self.numpy_metric is not None:
            out = np.empty(data.shape[1:] if data.ndim >1 else (1,))
            try:
                getattr(np, self.numpy_metric)(data, axis=0, out=out)
            except TypeError as e:
                raise RuntimeError(f'numpy metric: {self.numpy_metric} failed with {e}')
            except Exception as e:
                raise RuntimeError(f'Encountered unexpected error related to numpy metric {e}')
            data = out
        else:
            data = np.array(data) if copy else data

        do_compression = compress if self._compression_override is None else not self._compression_override

        # from line_profiler import LineProfiler
        # profile = LineProfiler()
        # profile.enable_by_count()
        times = []
        times.append(time.perf_counter())
        #        getLogger(__name__).debug(f'MiB Free: {memfree_mib()}')
        compressed = blosc2.compress(data) if do_compression else data
        times.append(time.perf_counter())
        #        getLogger(__name__).debug(f'MiB Free: {memfree_mib()}')
        self._send_status('capturing', status)
        times.append(time.perf_counter())
        tracker = self._data_socket.send_multipart([self.id, compressed], copy=False, track=not copy)
        times.append(time.perf_counter())
        # profile.disable_by_count()
        #getLogger(__name__).debug(profile.get_stats())
        times = np.diff(times) * 1000
        getLogger(__name__).debug(f'Compress {times[0]:.2f} ms, Status Up {times[1]:.2f}, Send {times[2]:.2f}')
        cval = 100 * len(compressed) / data.nbytes if data.nbytes else 100
        getLogger(__name__).debug(f'Sending {format_bytes(len(compressed))}, '
                                  f'compressed to {cval:.1f} % from {format_bytes(data.nbytes)}.')
        return tracker

    def _send_status(self, status, message=''):
        if not self._status_socket:
            raise RuntimeError('No status_socket connection available')
        update = f'{status}:{message}'
        getLogger(__name__).getChild('statusio').debug(f'Published status update for {self}: "{update}"')
        self._last_status = (status, message)
        self._status_socket.send_multipart([self.id, update.encode()])

    @property
    def completed(self):
        return self._last_status is not None and self._last_status[0] in ('finished', 'aborted', 'failed')

    def set_status(self, status, message='', context: zmq.Context = None, destablish=False):
        """
        get appropriate context and send current status message after connecting the status socket

        context is ignored if a status destination connection is extant
        """
        if not self._status_socket:
            context = context or zmq.Context().instance()
            self._status_socket = context.socket(zmq.PUB)
            self._status_socket.connect(self.STATUS_ENDPOINT)
        self._send_status(status, message)
        if destablish:
            self.destablish()

    @property
    def size_bytes(self):
        if self.numpy_metric:
            return self.capture_atom_bytes
        else:
            return self.nsamp*self.capture_atom_bytes

    @property
    def nchan(self):
        if self.channels:
            return len(self.channels)
        elif self.tap == 'postage':
            return N_POSTAGE_CHANNELS
        elif 'iq' in self.tap or 'phase' in self.tap:
            return N_CHANNELS
        else:
            return 1

    @property
    def dwid(self) -> int:
        """Data size of a capture sample in bytes"""
        if self.numpy_metric:
            return 2 * 8  ## numpy math returns float64 (8 bytes) * 2 for IQ
        if 'iq' in self.tap or self.tap in ('adc', 'postage'):
            return 4
        elif 'phase' in self.tap:
            return 2
        else:
            return PHOTON_DTYPE.itemsize

    @property
    def capture_atom_bytes(self) -> int:
        return self.dwid*self.nchan

    @property
    def buffer_shape(self):
        """ (nsamples, nchannels, 1|2)  """
        n = int(np.ceil(self.nsamp * MAXIMUM_DESIGN_COUNTRATE_PER_S / 1000)) if self.tap == 'photon' else self.nsamp

        if self.tap == 'postage':
            return n, PHOTON_POSTAGE_WINDOW_LENGTH + 1, 2
        elif 'iq' in self.tap or self.tap == 'adc':
            if self.numpy_metric:
                return self.nchan, 2
            else:
                return n, self.nchan, 2
        elif self.tap == 'photon':
            return (n,)
        elif 'phase' in self.tap:
            if self.numpy_metric:
                return self.nchan
            else:
                return n, self.nchan
        else:
            raise RuntimeError('Unknown tap: {self.tap}')


class CaptureSink(threading.Thread):
    def __init__(self, req_nfo: [CaptureRequest, tuple], start=True):
        if isinstance(req_nfo, CaptureRequest):
            id, buffer_shape, size_bytes, source_url, _compression_override = req_nfo.id, req_nfo.buffer_shape, req_nfo.size_bytes, req_nfo.server.data_url, req_nfo._compression_override
        else:
            id, buffer_shape, size_bytes, source_url, _compression_override = req_nfo
        super(CaptureSink, self).__init__(name=f'cap_id={id.decode()}')
        self._expected_buffer_shape = buffer_shape
        self._expected_bytes = size_bytes
        self.daemon = True
        self.cap_id = id
        self.data_source = source_url
        self.result = None
        self._pipe = None
        self._buf = []
        self._compression_override = _compression_override
        self.socket = None
        if start:
            self.start()

    def start(self):
        assert not self._started.is_set()
        self._pipe = zpipe(zmq.Context.instance())
        super(CaptureSink, self).start()

    def kill(self):
        self._pipe[0].send(b'')
        self._pipe[0].recv()
        self._pipe[1].close()
        self._pipe[0].close()

    def _accumulate_data(self, d):
        self._buf.append(d)

    def _finish_accumulation(self):
        self._buf = b''.join(self._buf)

    def _finalize_data(self):
        raise NotImplementedError

    def establish(self, ctx: zmq.Context = None):
        ctx = ctx or zmq.Context.instance()
        self.socket = ctx.socket(zmq.SUB)
        self.socket.setsockopt(zmq.SUBSCRIBE, self.cap_id)
        self.socket.connect(self.data_source)
        return self.socket

    def receive(self):
        id, data = self.socket.recv_multipart(copy=False)
        if not data:
            self.socket.close()
            return None
        return blosc2.decompress(data) if not self._compression_override else data

    def destablish(self):
        self.socket.close()
        self.socket = None

    def run(self):
        assert self.socket is None
        try:
            self.establish()
            poller = zmq.Poller()
            poller.register(self._pipe[1], flags=zmq.POLLIN)
            poller.register(self.socket, flags=zmq.POLLIN)

            getLogger(__name__).debug(f'Listening for data for {self.cap_id}')
            self._pipe[1].send(b'')
            while True:
                avail = dict(poller.poll())
                if self._pipe[1] in avail:
                    getLogger(__name__).debug(f'Received shutdown order, terminating data acq. of {self}')
                    break
                data = self.receive()  # does decompression
                if not data:
                    getLogger(__name__).debug(f'Received null, capture data stream over')
                    break
                getLogger(__name__).debug(f'Received {format_bytes(len(data))} of {format_bytes(self._expected_bytes)} for {self}')
                self._accumulate_data(data)
            self._finish_accumulation()
            self._finalize_data()
            if self.result:
                getLogger(__name__).info(f'Capture data for {self.cap_id} processed into {self.result.data.shape} '
                                         f'{self.result.data.dtype}: {self.result}')
            else:
                getLogger(__name__).info(f"Capture data for {self.cap_id} processed 'None'")

        except zmq.ZMQError as e:
            getLogger(__name__).warning(f'Shutting down {self} due to {e}')
        except AttributeError:
            raise
        finally:
            self.destablish()
            self._pipe[1].send(b'')

    def data(self, timeout=None):
        self.join(timeout=timeout)
        return self.result

    def ready(self):
        self._pipe[0].recv()  # block until a byte arrives


class StreamCaptureSink(CaptureSink):
    def _finalize_data(self):
        # raw adc data is i0q0 i1q1 int16
        bytes_recv = len(self._buf)
        if not bytes_recv:
            getLogger(__name__).warning(f'No data received for {self.cap_id}. '
                                        f'Expected {format_bytes(self._expected_bytes)}.')
            return
        elif bytes_recv > self._expected_bytes:
            getLogger(__name__).warning(f'Received more data than expected for {self.cap_id}. '
                                        f'Expected {format_bytes(self._expected_bytes)} got '
                                        f'{format_bytes(bytes_recv)}.')
        elif bytes_recv < self._expected_bytes:
            getLogger(__name__).warning(f'Finalizing incomplete capture data for {self.cap_id}. '
                                        f'Expected {format_bytes(self._expected_bytes)} got '
                                        f'{format_bytes(bytes_recv)}.')


        samples_recv = int(len(self._buf) / 2 // np.prod(self._expected_buffer_shape[1:]))  # / 2 for 2 bytes per I or Q
        shape = (min(samples_recv, self._expected_buffer_shape[0]),) + self._expected_buffer_shape[1:]

        dtype_bytes = self._expected_bytes // np.prod(shape, dtype=int)
        if dtype_bytes == 8:
            dtype = np.float64
        else:
            dtype = np.int16
            if dtype_bytes != 2:
                getLogger(__name__).warning(f'Received data has {dtype_bytes} bytes per sample. '
                                            f'Data may be lost in cast to int16 (2 bytes).')

        # TODO this might technically be a memory leak if we captured more data
        self.result = np.frombuffer(self._buf, count=np.prod(shape, dtype=int), dtype=dtype).reshape(shape).squeeze()


class ADCCaptureSink(StreamCaptureSink):
    def _finalize_data(self):
        super()._finalize_data()
        if self.result is not None:
            self.result = ADCCaptureData(self.result)


class IQCaptureSink(StreamCaptureSink):
    def _finalize_data(self):
        super()._finalize_data()
        if self.result is not None:
            self.result = IQCaptureData(self.result)


class PhaseCaptureSink(StreamCaptureSink):
    def _finalize_data(self):
        super()._finalize_data()
        if self.result is not None:
            self.result = PhaseCaptureData(self.result)


class SimplePhotonSink(CaptureSink):
    def _accumulate_data(self, d):
        self._buf.append(d)
        if sum(map(len, self._buf)) / 1024 ** 2 > 100:  # limit to 100MiB
            self._buf.pop(0)

    def _finish_accumulation(self):
        self._buf = b''.join(self._buf)

    def _finalize_data(self):
        self.result = PhotonCaptureData(self._buf)


class PhotonCaptureSink(CaptureSink):

    def capture(self, hdf, xymap, feedline_source, fl_ids):
        t = threading.Thread(target=self._main, args=(hdf, xymap, feedline_source, fl_ids))
        t.start()

    @staticmethod
    def _main(hdf, xymap, feedline_source, fl_ids, term_source='inproc://PhotonCaptureSink.terminator.inproc'):
        """

        Args:
            xymap: [nfeedline, npixel, 2] array
            feedline_source: zmq.PUB socket with photonbufers published by feedline
            term_source: a zmq socket of undecided type for detecting shutdown requests

        Returns: None

        """

        fl_npix = 2048
        n_fl = 5
        MAX_NEW_PHOTONS = 5000
        DETECTOR_SHAPE = (128, 80)
        fl_id_to_index = np.arange(n_fl, dtype=int)

        context = zmq.Context.instance()
        term = context.socket(zmq.SUB)
        term.setsockopt(zmq.SUBSCRIBE, id)
        term.connect(term_source)

        data = context.socket(zmq.SUB)
        data.setsockopt(zmq.SUBSCRIBE, fl_ids)
        data.connect(feedline_source)

        poller = zmq.Poller()
        poller.register(term, flags=zmq.POLLIN)
        poller.register(data, flags=zmq.POLLIN)

        live_image = np.zeros(DETECTOR_SHAPE)
        live_image_socket = None
        live_image_by_fl = live_image.reshape(n_fl, fl_npix)
        photons_rabuf = np.recarray(MAX_NEW_PHOTONS,
                                    dtype=(('time', 'u32'), ('x', 'u32'), ('y', 'u32'),
                                           ('phase', 'u16')))

        while True:
            avail = poller.poll()
            if term in avail:
                break

            frame = data.recv_multipart(copy=False)
            fl_id = frame[0]
            time_offset = frame[1]
            d = blosc2.decompress(frame[2])
            frame_duration = None  # todo time coverage of data
            # buffer is nchan*nmax+1 32bit: 16bit time(base2) 16bit phase
            # make array of to [nnmax+1, nchan, 2] uint16
            # nmax will always be <<2^12 number valid will be at [0,:,0]
            # times need oring w offset
            # photon data is d[1:d[0,i,0], i, :]

            nnew = d[0, :, 0].sum()
            # if we wanted to save binary data then we could save this, the x,y list, and the time offset
            # mean pixel count rate in this packet is simply [0,:,0]/dt
            fl_ndx = fl_id_to_index[fl_id]
            live_image_by_fl[fl_ndx, :] += d[0, :, 0] / frame_duration

            # if live_image_ready
            live_image_socket.send_multipart([f'liveim', blosc2.compress(live_image)])

            cphot = np.cumsum(d[0, :, 0], dtype=int)
            for i in range(fl_npix):
                sl_out = slice(cphot[i], cphot[i] + d[0, i, 0])
                sl_in = slice(1, d[0, i, 0])
                photons_rabuf['time'][sl_out] = d[sl_in, :, 0]
                photons_rabuf['time'][sl_out] |= time_offset
                photons_rabuf['phase'][sl_out] = d[sl_in, :, 1]
                photons_rabuf['x'][sl_out] = xymap[fl_ndx, i, 0]
                photons_rabuf['y'][sl_out] = xymap[fl_ndx, i, 1]
            hdf.grow_by(photons_rabuf[:nnew])

        term.close()
        data.close()
        hdf.close()


class PostageCaptureSink(CaptureSink):
    def _finalize_data(self):
        # get_postage returns:
        #  ids (nevents uint16 array or res cahnnels),
        #  events (nevents x 127 x2 of iq da

        if len(self._buf) != np.prod(self._expected_buffer_shape) * 4:
            getLogger(__name__).warning(f'Finalizing incomplete capture data for {self.cap_id}')
        self.result = PostageCaptureData(self._buf)


class StatusListener(threading.Thread):
    def __init__(self, id, source, initial_state='Created', start=True):
        super().__init__(name=f'StautsListner_{id.decode()}')
        self.daemon = True
        self.source = source
        self._pipe = None
        self.id = id
        self._status_messages = [initial_state]
        if start:
            self.start()

    def start(self):
        assert not self._started.is_set()
        self._pipe = zpipe(zmq.Context.instance())
        super(StatusListener, self).start()

    def kill(self):
        self._pipe[0].send(b'')
        self._pipe[0].recv()
        self._pipe[0].close()
        self._pipe[1].close()

    @staticmethod
    def is_final_status_message(update):
        """return True iff message is a final status update"""
        for r in CaptureRequest.FINAL_STATUSES:
            if update.startswith(r):
                return True
        return False

    def run(self):
        try:
            ctx = zmq.Context().instance()
            with ctx.socket(zmq.SUB) as sock:
                sock.linger = 0
                sock.setsockopt(zmq.SUBSCRIBE, self.id)
                sock.connect(self.source)

                poller = zmq.Poller()
                poller.register(self._pipe[1], flags=zmq.POLLIN)
                poller.register(sock, flags=zmq.POLLIN)
                getLogger(__name__).debug(f'Listening for status updates to {self.id}')
                self._pipe[1].send(b'')
                while True:
                    avail = dict(poller.poll())
                    if self._pipe[1] in avail:
                        getLogger(__name__).debug(f'Shutdown requested, terminating {self}')
                        break
                    elif sock not in avail:
                        time.sleep(.1)  # play nice
                        continue
                    id, update = sock.recv_multipart()
                    assert id == self.id or not self.id
                    update = update.decode()
                    self._status_messages.append(update)
                    getLogger(__name__).debug(f'Status update for {id}: {update}')
                    if self.is_final_status_message(update):
                        break
        except zmq.ZMQError as e:
            getLogger(__name__).critical(f"{self} died due to {e}")
        finally:
            self._pipe[1].send(b'')

    def latest(self):
        return self._status_messages[-1]

    def history(self):
        return tuple(self._status_messages)

    def ready(self):
        self._pipe[0].recv()  # TODO make this line block


class ADCCaptureData:
    """
    Formats raw ADC capture data to complex floats representing mV at the ADC input SMA.
    """

    def __init__(self, raw_data):
        """
        ADC data captured by an ADC capture request
        Args:
            raw_data: raw data captured by adc capture
        """
        self.raw = raw_data

    @property
    def dtype(self):
        return self.raw.dtype

    @property
    def shape(self):
        return self.raw.shape

    @cached_property
    def data(self):
        return convert_adc_raw_to_mv(self.raw[:, 0] + 1j * self.raw[:, 1])


class IQCaptureData:
    """
    Formats raw IQ capture data to complex floats between +/- 1.
    """

    def __init__(self, raw_data):
        self.raw = raw_data

    @property
    def dtype(self):
        return self.raw.dtype

    @property
    def shape(self):
        return self.raw.shape

    @cached_property
    def data(self):
        return raw_iq_to_unit(self.raw[..., 0] + self.raw[..., 1] * 1j)


class PhaseCaptureData:
    def __init__(self, data):
        self.raw = data

    @property
    def dtype(self):
        return self.raw.dtype

    @property
    def shape(self):
        return self.raw.shape

    @cached_property
    def data(self):
        return raw_phase_to_radian(self.raw, scaled=False)


class PhotonCaptureData:
    def __init__(self, buf):
        self.raw = np.frombuffer(buf, dtype=PHOTON_DTYPE_PACKED)
        self.photons = unpack_photons(self.raw)

    @property
    def data(self):
        return self.photons

    @property
    def dtype(self):
        return PHOTON_DTYPE

    @property
    def shape(self):
        return self.photons.size


class PostageCaptureData:
    def __init__(self, buf) :
        result = np.frombuffer(buf, count=len(buf) // 2, dtype=np.int16)
        n = result.size//2//128
        shape = (n, 128, 2)
        result = result.reshape(shape)
        from ..funcs import postage_buffer_to_data
        ids, events = postage_buffer_to_data(result, complex=True, scaled=True)
        self.ids = ids  # result[:, 0, 0].astype(np.uint16)
        self.iqs = events  # result[:, 1:, :]
        self.raw = result

    @property
    def dtype(self):
        return np.int16

    @property
    def shape(self):
        return self.iqs.shape

    @property
    def data(self):
        return self.iqs


class CaptureJob:
    def __init__(self, request: CaptureRequest, submit=False):
        """
        Args:
            request:
            submit: tuple of bools (with sink?, with status?) or True or False
        """
        self.request = request
        self.submitted = False

        self._status_listener = StatusListener(request.id, request.server.status_url,
                                               initial_state='created', start=False)

        if request.tap == 'adc':
            datasink = ADCCaptureSink
        elif 'iq' in request.tap:
            datasink = IQCaptureSink
        elif 'phase' in request.tap:
            datasink = PhaseCaptureSink
        elif request.tap == 'photon':
            datasink = SimplePhotonSink
        elif request.tap == 'postage':
            datasink = PostageCaptureSink
        else:
            raise ValueError(f'Unknown tap location {request.tap}')

        self.datasink = datasink(request, start=False)

        if submit:
            self.submit()
        elif isinstance(submit, tuple):
            self.submit(**submit)

    def __del__(self):
        self._kill_workers(kill_status_monitor=True, kill_data_sink=True)

    def status(self):
        """ Return the last known status of the request """
        return self._status_listener.latest()

    def status_history(self) -> tuple[str]:
        """Return a tuple of the status history of the job"""
        return self._status_listener.history()

    def cancel(self, kill_status_monitor=False, kill_data_sink=True):
        """

        Args:
            kill_status_monitor: Whether to terminate the status monitor
            kill_data_sink: Whether to terminate the data saver
        Returns: The parsed json submission result

        """
        self._kill_workers(kill_status_monitor=kill_status_monitor, kill_data_sink=kill_data_sink)
        ctx = zmq.Context().instance()
        with ctx.socket(zmq.REQ) as s:
            s.connect(self.request.server.command_url)
            s.send_pyobj(('abort', self.request.id))
            return s.recv_pyobj()

    def _kill_workers(self, kill_status_monitor=True, kill_data_sink=True):
        if kill_status_monitor and self._status_listener.is_alive():
            try:
                self.datasink.kill()
            except zmq.ZMQError as e:
                getLogger(__name__).warning(f'Caught {e} when telling data sink to terminate')
            if self._status_listener.is_alive():
                getLogger(__name__).warning(f'Status listener did not instantly terminate')
        if kill_data_sink and self.datasink.is_alive():
            try:
                self.datasink.kill()
            except zmq.ZMQError as e:
                getLogger(__name__).warning(f'Caught {e} when telling data sink to terminate')
            if self.datasink.is_alive():
                getLogger(__name__).warning(f'Data sink did not instantly terminate')

    def data(self, timeout=None):
        return self.datasink.data(timeout=timeout)

    def submit(self, with_sink=False, with_status=True):
        if with_status:
            self._status_listener.start()
            self._status_listener.ready()

        if with_sink and not self.request.data_endpoint.startswith('file://'):
            self.datasink.start()
            self.datasink.ready()

        try:
            ret = self.request.submit()
            self.submitted = True
            return ret
        except Exception as e:
            self._status_listener.kill()
            self.datasink.kill()
            raise e



import enum

from typing import Type, Optional


def getmask(width: int):
    """Get an int with the low `width` bits set

    Parameters
    ----------
    width : int
        The number of bits to set in the output
    """
    return (1 << width) - 1


def checkslice(sl: slice, width: int):
    """Checks that a slice is valid for a particular register width

    Parameters
    ----------
    sl : slice
        The slice to check
    width : int
        The underlying register width
    """
    if not (sl.step is None) and sl.stride != 1:
        raise IndexError("Access Stride must be 1 or None")
    if sl.stop > width or sl.start >= width:
        raise IndexError("Access wider than register width ({:d}".format(width))
    if sl.stop < sl.start:
        raise IndexError("Reverse indexing unsupported")
    if sl.stop < 0 or sl.start < 0:
        raise IndexError("Negative indicies unsupported")


class MetaRegister:
    """A metaclass for both `Register` and `Field`"""

    _objcache = None

    def __get__(self, obj, objtype=None):
        pass

    def __set__(self, obj, val):
        pass

    def __len__(self):
        pass

    def __getitem__(self, obj, sl: int | slice):
        """Extract a field from a register

        Get a slice of this register, this is only callable from subclasses
        of this class, if you want to extract a field on the fly you can do
        that with:

        ```python
        reg = Register(...)
        field_data = field(slice(0, 4), reg).__get__(self)
        ```

        Where `self` is a reference to the IP Object

        Parameters
        ----------
        obj
            A reference to the underlying IP
        sl : int | slice
            The bit(s) to access

        Returns
        -------
        int : the requested field
        """
        self._objcache = obj
        if type(sl) is int:
            val = self.__get__(obj)
            return (val >> sl) & 1
        checkslice(sl, self.__len__())
        if sl.start is None:
            return self.__getitem__(obj, self.slice(0, sl.stop))
        if sl.stop == sl.start:
            return self.__getitem__(obj, sl.start)
        mask = getmask(sl.stop - sl.start)
        val = self.__get__(obj)
        return mask & (val >> sl.start)

    def __setitem__(self, obj, sl: int | slice, val: int):
        """Sets a field to a given value, see `MetaRegister.__get__`

        Parameters
        ----------
        obj
            A reference to the underlying IP
        sl : int | slice

        """
        self._objcache = obj
        if type(sl) is int:
            val_init = self.__get__(obj)
            self.__set__(obj, (val_init & (getmask(32) ^ (1 < sl))) | (val << sl))
            return
        checkslice(sl, self.__len__())
        if sl.start is None:
            return self.__setitem__(obj, self.slice(0, sl.stop), val)
        if sl.stop == sl.start:
            return self.__setitem__(obj, self.slice(0, sl.stop), val)
        mask = getmask(sl.stop - sl.start)
        val_init = self.__get__(obj)
        self.__set__(
            obj,
            (val_init & (getmask(32) ^ (mask << sl.start))) | val << sl.start,
        )


class Register(MetaRegister):
    """A read write register referencing an underlying PYNQ register

    This exists to make building pynq IP marginally more ergonomic, it allows
    easy creation of fields at runtime that don't exist in an HWH description

    When accessing the underlying register in an IP it should always look like
    an int, accessing always reads and assignments always write, methods other
    than `__get__` and `__set__` are typically only available to fields of this
    register

    Typically instead of instantiating this directly you will want to use the
    `register` decorator on a method in your class, if you have an underlying
    register in the HWH description for an IP named `my_reg` you would
    instantiate this like:

    ```python
    import registers

    class MyIP(pynq.DefaultIP):
        ...
        @registers.register
        def my_reg(self):
            return self.register_map.my_reg
    ```

    This slightly odd API exists because the registers can't be resolved
    until the class is instantiated and the HWH is parsed
    """

    def __init__(self, getreg):
        """
        Initilize the class with a function that returns the underlying pynq
        register, this function will recieve one argument, a reference to the
        underlying IP which this register is an attribute of

        If you would like to instantiate this with a register not contained
        in the HWH file you can do at some offeset `OFFSET`:
        ```python
        my_custom_reg = registers.Register(lambda self: pynq.Register(OFFSET, buffer=self.mmio.array))
        ```

        Which would be equivalent to:
        ```python
        class MyIP(DefaultIP):
            ...
            @registers.register
            def my_custom_reg(self):
                return pynq.Register(OFFSET, buffer=self.mmio.array)
        ```

        Arguments
        ---------
        getreg : func(self)
            A function which is passed an argument to the enclosing IP and returns
            a pynq register
        """
        self.getreg = getreg

    def __len__(self):
        """Get the width of this register in bits"""
        # Return a sensible default, maybe should error
        if self._objcache is None:
            return 32
        return self.getreg(self._objcache).width

    def __get__(self, obj, objtype=None):
        """Read the current value of this register"""
        self._objcache = obj
        if obj is None or ((objtype is not None) and issubclass(objtype, MetaRegister)):
            return self
        return self.getreg(obj)[:]

    def __set__(self, obj, val: int):
        """Set the Value of this register"""
        self._objcache = obj
        self.getreg(obj)[:] = val


class RegisterRO(Register):
    """A read only register, can be created with the decorator `register_ro`"""

    def __set__(self, obj, val: int):
        raise ValueError("Attempted to write read only register")


class RegisterWO(Register):
    """A write only register, can be created with the decorator `register_wo`"""

    def __get__(self, obj, objtype=None):
        raise ValueError("Attempted to read write only register")


class RegisterShadow(Register):
    """A write only register with a shadow backing store

    Writes write to the underlying register, reads read the last value written
    and error if there haven't been writes since instantiation

    This can be created with decorator `register_shadow` which can also be used
    to instantiate a version of this class with a default value
    """

    _shadow = None

    def __set__(self, obj, val: int):
        self._shadow = val
        super().__set__(obj, val)

    def __get__(self, obj, objtype=None):
        if self._shadow is None:
            raise ValueError("Shadow register not initilized and never written to")
        return self._shadow


def register(getreg):
    """A decorator which can be used to create a read write register

    For usage see the documentation for `Register`

    Parameters
    ----------
    getreg : func(self)
        A function which takes a reference to the enclosing IP and returns a pynq register
    """
    return Register(getreg)


def register_ro(getreg):
    """See `register`"""
    return RegisterRO(getreg)


def register_wo(getreg):
    """See `register`"""
    return RegisterWO(getreg)


def register_shadow(arg):
    """A decorator which creates an optionally initilized shadow register

    When used like the other decorators this produces an uninit shadow register.

    To create a shadow register initilized to some value (in this case 128):
    ```python
    class MyIP(DefaultIP):
        ...
        @registers.register_shadow(128)
        def my_shadow_reg(self):
            return self.registers.my_write_only_register
    ```
    """
    if type(arg) is int:

        class RegisterShadowInit(RegisterShadow):
            _shadow = arg

        return lambda getreg: RegisterShadowInit(getreg)
    return RegisterShadow(arg)


class Field(MetaRegister):
    """A class representing an individual register field

    The backing parent can be a mkidgen3 `Register` or any other concerete
    subclass of `MetaRegister`. The recommended way to create this is to
    use the `field` decorator.

    There are two ways to approach this, the recommended way exists for
    consistency with the `Register` interface:

    ```python
    import mkidgen3.registers as registers

    class MyIP(DefaultIP):
        ...
        @registers.register_ro
        def status_register(self):
            return self.register_map.status_register
        ...
        # Interrupt Count is bits [0-7] inclusive of the status register
        @registers.field(slice(0, 8))
        def interrupt_count(self):
            return self.status_register
        ...
    ```

    Because mkidgen3 style registers will typically be definition-time
    resolvable, the alternate invocation style is:

    ```python
    class MyIP(DefaultIP):
        ...
        interrupt_count = field(slice(0, 8), status_register)
        ...
    ```
    """

    def __init__(self, parentfunc, sl: slice | int):
        self.parentfunc = parentfunc
        self.sl = sl

    def __get__(self, obj, objtype=None):
        self._objcache = obj
        return self.parentfunc(obj).__getitem__(obj, self.sl)

    def __set__(self, obj, val):
        self._objcache = obj
        self.parentfunc(obj).__setitem__(obj, self.sl, val)


class FieldEnum(Field):
    """This is a field representing an one of several enumerated options

    This is handy in many cases when a register represents one of several
    options, for example:

    ```
    import mkidgen3.registers as registers
    from enum import Enum

    class CaptureSource(Enum):
        NONE = 0
        PHASE = 1
        IQ = 2
        RAW_PHASE = 3

    class MyIP(DefaultIP):
        ...
        @registers.register_ro
        def status_register(self):
            return self.register_map.status_register
        ...
        # The captue source is bits [8:10] inclusive of the status register
        capture_source = field_enum(slice(8, 11), CaptureSource, status_Register)
        ...
    ```
    """

    def __init__(self, parentfunc, sl: slice | int, enum: Type[enum.Enum]):
        super().__init__(parentfunc, sl)
        self.enum = enum

    def __get__(self, obj, objtype=None):
        self._objcache = obj
        return self.enum(super().__get__(obj, objtype))

    def __set__(self, obj, val):
        self._objcache = obj
        if not type(val) is self.enum:
            raise TypeError(
                "Value passed must be of type {:s}, got {:s} ({:s})".format(
                    repr(self.enum), repr(val), repr(type(val))
                )
            )
        super().__set__(obj, val.value)


class FieldBool(Field):
    """This is a single bit field representing a boolean

    To instantiate this you typically want the `field_bool` decorator/factory func
    ```
    import mkidgen3.registers as registers
    ...
    class MyIP(DefaultIP):
        ...
        @registers.register_ro
        def status_register(self):
            return self.register_map.status_register
        ...
        # Bit 11 of the status register indicates if the active capture has completed
        capture_complete = field_bool(11, status_register)
        ...
    ```
    """

    def __init__(self, parentfunc, sl: int):
        super().__init__(parentfunc, sl)

    def __get__(self, obj, objtype=None):
        return bool(super().__get__(obj, objtype))

    def __set__(self, obj, val):
        super().__set__(obj, int(val))


def field(bits: slice | int, reg: Optional[Type[MetaRegister]] = None):
    """Decorator for creating a field from a register and a slice, see `Field`"""
    if reg is not None:
        return Field(lambda _: reg, bits)
    return lambda regfunc: Field(regfunc, bits)


def field_enum(bits: slice | int, enum, reg: Optional[Type[MetaRegister]] = None):
    """Decorator for creating a field from a register, slice, and enum, see `FieldEnum`"""
    if reg is not None:
        return FieldEnum(lambda _: reg, bits, enum)
    return lambda regfunc: FieldEnum(regfunc, bits, enum)


def field_bool(bit: int, reg: Optional[Type[MetaRegister]] = None):
    """Decorator for creating a boolean field from a register and bit index, see `FieldBool`"""
    if reg is not None:
        return FieldBool(lambda _: reg, bit)
    return lambda regfunc: FieldBool(regfunc, bit)



import logging

import numpy as np
import numpy.typing as nt
import scipy.optimize as opt

from typing import Optional, Callable

from mkidgen3.server.feedline_config import DDCConfig
from mkidgen3.drivers.ddc import ThreepartDDC
from mkidgen3.mkid_setup.sweeps import Sweep


def pointslope_intersect(
    m0, x0, y0, m1, x1, y1, m0r=0, x0r=0, y0r=0, m1r=0, x1r=0, y1r=0
):
    xin = y0 - y1 + m1 * x1 - m0 * x0
    xid = m1 - m0
    xi = xin / xid

    yip = xi - x0
    yi = y0 + m0 * yip

    xirn = np.sqrt(
        y0r**2
        + y1r**2
        + (m1 * x1 * np.sqrt((m1r / m1) ** 2 + (x1r / x1) ** 2)) ** 2
        + (m0 * x0 * np.sqrt((m0r / m0) ** 2 + (x0r / x0) ** 2)) ** 2
    )
    xird = np.sqrt(m1r**2 + m0r**2)
    xir = xi * np.sqrt((xirn / xin) ** 2 + (xird / xid) ** 2)
    yirp = np.sqrt(xir**2 + x0r**2)
    yir = np.sqrt(y0r**2 + ((m0r / m0) ** 2 + (yirp / yip) ** 2) * ((yip * m0) ** 2))

    return xi, yi, xir, yir


def slope_midpoint(x0, x1, y0, y1, x0r, x1r, y0r, y1r):
    md = x1 - x0
    mn = y1 - y0

    mnr = np.sqrt(y1r**2 + y0r**2)
    mdr = np.sqrt(x1r**2 + x0r**2)

    m = mn / md
    mr = m * np.sqrt((mnr / mn) ** 2 + (mdr / md) ** 2)

    ymidr = mnr / 2
    xmidr = mdr / 2

    xmid = (x1 + x0) / 2
    ymid = (y1 + y0) / 2

    return m, mr, xmid, ymid, xmidr, ymidr


def approximate_center(i, q, ir, qr, plot=None, holdoff=6):
    assert i.size == q.size

    mi = q.size // 2

    cestimates = np.zeros((q.size // 2 - 1 - holdoff, 2))
    cestimaters = np.zeros((q.size // 2 - 1 - holdoff, 2))

    for p in range(holdoff, q.size // 2 - 1):
        mp, mpr, xmp, ymp, xmpr, ympr = slope_midpoint(
            i[mi], i[mi + p], q[mi], q[mi + p], ir[mi], ir[mi + p], qr[mi], qr[mi + p]
        )
        mn, mnr, xmn, ymn, xmnr, ymnr = slope_midpoint(
            i[mi], i[mi - p], q[mi], q[mi - p], ir[mi], ir[mi - p], qr[mi], qr[mi - p]
        )
        mpi, mpir = -1 / mp, mpr / (mp**2)
        mni, mnir = -1 / mn, mnr / (mn**2)
        xc, yc, xcr, ycr = pointslope_intersect(
            mpi, xmp, ymp, mni, xmn, ymn, mpir, xmpr, ympr, mnir, xmnr, ymnr
        )
        cestimates[p - holdoff][0] = xc
        cestimates[p - holdoff][1] = yc
        cestimaters[p - holdoff][0] = xcr
        cestimaters[p - holdoff][1] = ycr

    if plot:
        plot.errorbar(
            cestimates[::, 0], cestimates[::, 1], cestimaters[::, 0], cestimaters[::, 1]
        )

    cestimate = np.sum(cestimates / (cestimaters**2), axis=0) / np.sum(
        1 / (cestimaters**2), axis=0
    )

    return (cestimate[0], cestimate[1]), np.sqrt(
        (cestimate[0] - i[mi]) ** 2 + (cestimate[1] - q[mi]) ** 2
    )


def circle_model(i, r, i0, q0):
    return np.sqrt(r**2 - (i - i0) ** 2) + q0


def find_truecenter(
    i,
    q,
    irms,
    qrms,
    center,
    radius,
    model=circle_model,
    plot=None,
):
    i_apcent = i - center[0]
    q_apcent = q - center[1]

    angles = np.arctan2(q_apcent, i_apcent)
    middle = (angles[0] + angles[-1]) / 2
    rotation = np.pi / 2 - middle

    i_apcentrot = i_apcent * np.cos(rotation) - q_apcent * np.sin(rotation)
    q_apcentrot = i_apcent * np.sin(rotation) + q_apcent * np.cos(rotation)

    if plot:
        plot.plot(i_apcentrot, q_apcentrot)
    popt, pcov = opt.curve_fit(
        model,
        i_apcentrot,
        q_apcentrot,
        sigma=np.sqrt(irms**2 + qrms**2),
        p0=(radius, 0, 0),
    )
    true_center = (
        center[0] + popt[1] * np.cos(-rotation) - popt[2] * np.sin(-rotation),
        center[1] + popt[1] * np.sin(-rotation) + popt[2] * np.cos(-rotation),
    )
    true_radius = popt[0]
    return true_center, true_radius


def quantize_rotation(rotation, bits=8):
    if rotation == np.pi:
        rotation = -np.pi
    return np.floor(rotation * (1 << (bits - 1)) / np.pi) * np.pi / (1 << (bits - 1))


def rotate_and_center(
    sweep: Sweep,
    targets: Optional[list[float] | nt.NDArray[np.float64]] = None,
    fit: Optional[Callable[..., nt.NDArray[np.float64]]] = None,
    program: Optional[ThreepartDDC] = None,
    plot: bool = False,
):
    if not targets:
        targets = np.zeros_like(sweep.iq.shape[0])

    phase_offsets = np.zeros(2048)
    ddc_centers = np.zeros(2048, dtype=np.complex64)

    iqs, iqrs = sweep.iq, sweep.iqsigma
    if iqrs is None:
        iqrs = np.ones_like(iqs)

    for t in range(len(targets)):
        i = iqs[t].real
        q = iqs[t].imag
        if plot:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots(1)
            ax.set_aspect("equal")
            ax.set_title(str(sweep.config.waveform.default_ddc_config.tones[t]))
            ax.plot(i, q)

        ri = iqrs[t].real
        rq = iqrs[t].imag

        target = targets[t]

        approxc, approxr = approximate_center(i, q, ri, rq, plot=ax if plot else None)
        if fit:
            center, radius = find_truecenter(
                i, q, ri, rq, approxc, approxr, plot=ax if plot else None
            )
        else:
            center = approxc
            radius = approxr

        logging.getLogger(__name__).debug(
            "Tone {:.4f} @ I{:.3f}xQ{:.3f} R{:.3f}".format(
                sweep.config.waveform.default_ddc_config.tones[t],
                center[0],
                center[1],
                radius,
            )
        )

        if plot:
            ax.add_patch(plt.Circle(approxc, approxr, color="b", fill=False))
            ax.add_patch(plt.Circle(center, radius, color="r", fill=False))
            ax.scatter(*center)

        current_rotation = np.arctan2(
            q[q.size // 2] - center[1], i[i.size // 2] - center[0]
        )
        rotation = quantize_rotation(target - current_rotation)
        rotated_center = (
            np.cos(rotation) * center[0] - np.sin(rotation) * center[1],
            np.sin(rotation) * center[0] + np.cos(rotation) * center[1],
        )

        phase_offsets[t] = rotation
        rotation = ((rotation + np.pi) % (2 * np.pi)) - np.pi
        ddc_centers[t] = rotated_center[0] + rotated_center[1] * 1j
        if plot:
            plt.show()

    ddc_centers /= 2**15

    ddc_config = DDCConfig(
        tones=None,
        phase_offset=phase_offsets,
        loop_center=ddc_centers,
    )
    if program:
        program.configure(**ddc_config.settings_dict())
    return ddc_config, center, radius



import logging

import numpy as np
from hashlib import md5
import copy
from mkidgen3.util import convert_freq_to_ddc_bins
from logging import getLogger





def _hasher(v, pass_none=False):
    """
    Hash a value

    Args:
        v: the value to be hashed, the md5 hexdigest of the bytestring of v is used. .tobytes() will be used if
        present otherwise str().encode() is used. hash() is called on _FLConfigMixin or _FLMetaonfigMixin objects.
        pass_none: If true, None is passed through without hashing, otherwise the string '___python_None' is
        used in its place.

    Returns: The hash
    """
    if v is None:
        return None if pass_none else md5(str('___python_None').encode()).hexdigest()
    elif isinstance(v, (_FLConfigMixin, _FLMetaconfigMixin)):
        return hash(v)
    else:
        try:
            v = v.tobytes()
        except AttributeError:
            v = str(v).encode()
        return md5(v).hexdigest()


class _FLConfigMixin:
    _settings = tuple()

    def __eq__(self, other):
        """
        Configs are equivalent if all of their settings are the equ.

        If either is hashed then their hashes must match exactly. Otherwise the data

        """
        if other is None:
            return self.empty()
        if not isinstance(other, type(self)):
            return False
        if self.is_hashed or other.is_hashed:
            return hash(self) == hash(other)
        for (_, a), (_, b) in zip(self._hash_data, other._hash_data):
            if not self._hashdata_vals_equal(a, b):
                return False
        return True

    def __ge__(self, other):
        """ Returns true if self is at least as specified as other (i.e. other has = or more Nones)"""
        if other is None:
            return True
        if not isinstance(other, type(self)):
            raise TypeError(f'Invalid type: {type(other)}')
        if self.is_hashed or other.is_hashed:
            raise ValueError('Unable to compare hashed configs.')
        for (_, a), (_, b) in zip(self._hash_data, other._hash_data):
            if b is not None and a is None:
                return False
        return True

    def __le__(self, other):
        if other is None:
            return self.empty()
        if not isinstance(other, type(self)):
            raise TypeError('Incorrect type for config comparison')
        return other.__ge__(self)

    def __lt__(self, other):
        """ Returns true if other is more specified than self (i.e. self has more Nones) """
        return not self.__ge__(other)

    def __gt__(self, other):
        return self.__ge__(other) and not self == other

    def __bool__(self):
        """ True iff not empty i.e. specifies config settings """
        return not self.empty()

    def __hash__(self):
        """
        Return a hash of the config, the same settings will always return the same hash.

        A hashed config has the same hash as an unhashed config.

        A setting set to None is different than one set to a value.
        """
        if self.is_hashed:
            return self._hashed
        return int(_hasher(self._hash_data), 16)

    def __str__(self):
        """Nicely format one's self"""
        name = self.__class__.__name__
        if self._hashed:
            return f"{name}: {hash(self)} (hashed)"
        else:
            return (f"{name}: {hash(self)}\n"
                    f"  {self.settings_dict()}")

    @staticmethod
    def _hashdata_vals_equal(a, b) -> bool:
        """Compare settings values for equality"""
        if type(a) != type(b):
            return False
        if isinstance(a, np.ndarray) and isinstance(b, np.ndarray) and not np.allclose(a, b):
            return False
        elif a != b:
            return False
        return True

    @property
    def _hash_data(self) -> tuple:
        if self.is_hashed:
            raise RuntimeError('Not supported for hashed FLConfig')
        """Return a tuple of (setting_key, hashed_data|None) pairs sorted on the setting key"""
        hash_data = ((k, _hasher(getattr(self, k), pass_none=True)) for k in self._settings)
        return tuple(sorted(hash_data, key=lambda x: x[0]))

    def merge_with(self, other):
        """Adopt the specified values of the other"""
        if other is None:
            return
        assert isinstance(other, type(self)), 'other must be of the same type'
        assert not self.is_hashed and not other.is_hashed, 'Hashed FLConfigs can not be merged'
        for k in self._settings:
            v = getattr(other, k)
            if v is not None:
                setattr(self, k, v)

    def compatible_with(self, other) -> bool:
        """
        Determine if two FLConfig's settings are compatible with one another.

        The present implementation is maximally conservative: all settings much match, None matches anything
        (see _hashdata_vals_equal for details of equality)
        Args:
            other: Another FLConfig

        Returns: True if the two configs' settings can coexist in hardware
        """
        if other is None:
            return True
        if not isinstance(other, type(self)):
            return False
        if other.is_hashed:
            if self.empty():
                return True
            else:
                return hash(self) == hash(other)
        for (_, a), (_, b) in zip(self._hash_data, other._hash_data):
            if a is None or b is None:
                continue
            if not self._hashdata_vals_equal(a, b):
                return False
        return True

    def deltafy(self, other):
        """
        Generate a new config with only the settings that are specified by other but not self, or have changed in other
        from self, all other settings are None.

        """
        d = {}
        if other is not None:
            for k in self._settings:
                sv = getattr(self, k)
                ov = getattr(other, k)
                if ov is not None:
                    if sv is None:
                        d[k] = ov
                        continue
                    if isinstance(ov, np.ndarray):
                        try:
                            if not (sv==ov).all():
                                d[k] = ov
                        except ValueError:
                            getLogger(__name__).warning(f'Array data comparison issue in {type(self)}.{k}.'
                                                        f'Check config definitions.')
                            d[k] = ov
                    elif sv != ov:
                        d[k] = ov
        return type(self)(**d)

    def empty(self):
        """
        Return true if all the settings are unspecified (e.g. None)
        """
        if self.is_hashed:
            return hash(self) == hash(type(self)())
        return not self.settings_dict(omit_none=True)

    @property
    def is_hashed(self) -> bool:
        return self._hashed is not None

    @property
    def hashed_form(self):
        """
        Return a new config of the same type in hashed form

        In hashed form the settings values are used to compute a hash, and it is stored instead of the settings values.
        """
        return type(self)(_hashed=hash(self))

    def unhashed_form(self, hash_cache: dict):
        if not self.is_hashed:
            return self
        d = self.settings_dict(unhasher_cache=hash_cache, _hashed=False, omit_none=True)
        return type(self)(**d)

    def settings_dict(self, omit_none=True, _hashed=False, unhasher_cache=None) -> dict:
        """
        Build a dict of setting_keys:values for use in creating a clone of the config or
        passing to a drivers configure() method

        TODO how does a hanshed None setting (is this possible?) interact with omit_none

        Args:
            omit_none: Exclude settings with no set value from the dictionary
            _hashed: for internal use building a hashed FLConfig, if True omit_none and unhasher_cache are ignored
            unhasher_cache: May be set to a dictionary to fetch actual values in liu of the hash.
                Required if calling and self.is_hashed is true. hash(self) is expected to be in the unhasher_cache.

        Returns: A dictionary of setting_key:value pairs, keys that are none may or may not be present.

        Raises: AssertionError if called on a hashed config without and unhasher_cache containing the hash.
        Will not be raised if _hashed is set.
        """
        if self._hashed and not _hashed:
            assert isinstance(unhasher_cache, dict), ('An unhasher_cache is required for to build a settings '
                                                      'dictionary from a hashed FLConfig')
            assert self._hashed in unhasher_cache, ('A hash entry in the unhasher_cache is required for to build a '
                                                    'settings dictionary from a hashed FLConfig')
            x = unhasher_cache[self._hashed]
        else:
            x = self
        if _hashed:
            d = {'_hashed': hash(self)}
        else:
            d = {k: copy.deepcopy(getattr(x, k)) for k in self._settings if not (getattr(x, k) is None and omit_none)}
        return d


class _FLMetaconfigMixin:
    """
    Like FLConfig the FLMetaconfig is a collection of settings for the PL and FL hardware (i.e. inclusive of
    clocking, and IF boards).

    All non-internal attributes are assumed to be instances of FLMetaConfigMixin or FLConfigMixin
    """

    def compatible_with(self, other):
        """
        Determine if two FLConfig's settings are compatible with one another.

        The present implementation is maximally conservative: all settings much match, None matches anything
        Args:
            other: Another FLConfig

        Returns: True if the two configs' settings can coexist in hardware
        """
        if other is None:
            return True
        if not isinstance(other, type(self)):  #NB a hased config is still a config, not an int for a metaconfig
            return False
        for k, v in self:
            other_v = getattr(other, k)
            assert isinstance(v, (_FLMetaconfigMixin, _FLConfigMixin, type(None), int))
            assert isinstance(other_v, (_FLMetaconfigMixin, _FLConfigMixin, type(None), int))
            if v is None or other_v is None:
                continue
            if not v.compatible_with(other_v):
                return False
        return True

    def __eq__(self, other):
        """
        Feedline configs are == if all of their settings are the same.
        NB: None is compatible with anything but not considered the same however None is == a config with all None settings
        """
        for k, v in self:
            other_v = None if other is None else getattr(other, k)
            assert isinstance(v, (_FLConfigMixin, type(None)))
            assert isinstance(other_v, (_FLConfigMixin, type(None)))
            if v is None and other_v is not None:
                if not other_v.empty():
                    return False
            elif not v == other_v:
                return False
        return True

    def __ge__(self, other):
        """Self is more specified or equal to other """

        for k, v in self:
            ov = getattr(other, k)
            if v is None and ov is not None and not ov.empty():
                return False
            if not v >= ov:
                return False
        return True

        # for k, v in self.iter():
        #     ov = getattr(other, k)
        #     if v is None and ov is not None:
        #         return False
        #     if v is None and ov is None:
        #         continue
        #     if not v >= ov:
        #         return False
        # return True

    def __le__(self, other):
        if other is None:
            return self.empty()
        if not isinstance(other, type(self)):
            raise TypeError('Incorrect type for config comparison')
        return other.__ge__(self)

    def __lt__(self, other):
        return not self.__ge__(other)

    def __gt__(self, other):
        return self.__ge__(other) and not self == other

    def __hash__(self):
        return int(_hasher(tuple(sorted(((k, _hasher(v)) for k, v in self.__dict__.items()), key=lambda x: x[0]))), 16)

    def __bool__(self):
        """ True iff not empty i.e. specifies config settings """
        return not self.empty()

    def __iter__(self):
        """
        An iterator of config_key: value pairs
        """
        for v in sorted(vars(self)):
            if v.startswith('_'):
                continue
            yield v, getattr(self, v)

    def empty(self):
        for _, v in self:
            if v is None or v.empty():
                continue
            return False
        return True

    def merge_with(self, other):
        """Adopt the specified values of the other"""
        if other is None:
            return
        assert isinstance(other, type(self)), 'other must be of the same type'
        for (_, v1), (_, v2) in zip(self, other):
            v1.merge_with(v2)

    @property
    def hashed_form(self):
        """
        Return a new config of the same type in hashed form

        In hashed form setting's that are FLConfigs are converted to their hashed form. This is done recursively
        through any keys that are FLMetaconfigs.
        """
        d = {k: v if v is None else v.settings_dict(_hashed=True) for k, v in self}
        return type(self)(**d)

    @property
    def is_hashed(self):
        for _, v in self:
            if v is not None and v.is_hashed:
                return True
        return False

    def unhashed_form(self, hash_cache: dict):
        if not self.is_hashed:
            return self
        d = self.settings_dict(unhasher_cache=hash_cache, _hashed=False, omit_none=True)
        return type(self)(**d)

    def settings_dict(self, _hashed=False, unhasher_cache=None, omit_none=True):
        """
        Build a dict of setting_keys:values for use in creating a clone of the config or
        passing to a drivers configure() method. The values of any settings that are FLMetaconfigs will
        themselves be dictionaries.

        Args:
            omit_none: Exclude settings with no set value from the dictionary
            _hashed: for internal use building a hashed FLConfig, if True omit_none and unhasher_cache are ignored
            unhasher_cache: May be set to a dictionary to

        Returns: A dictionary of setting_key:value pairs, values that are None may or may not be present. Values that
        are FLConfigs will be dictionaries
        """
        skip_none = omit_none
        if _hashed:
            skip_none = False
        return {k: v if v is None else v.settings_dict(omit_none=omit_none, _hashed=_hashed,
                                                       unhasher_cache=unhasher_cache)
                for k, v in self if not (v is None and skip_none)}

    def iter(self, hashed=True, unhashed=True):  # -> (str,FLConfigMixin|None):
        """
        An iterator of config_key: value pairs including any nested meta configs

        nested meta config's config keys are yielded as key.child_key, note that nothing prevents a child
        from also having meta config so.this.is.a.possible.key

        for example 'pp_config.trig_config', TrigerConfig|None
        """
        for k, v in self:
            if isinstance(v, _FLMetaconfigMixin):
                for a, b in v.iter(hashed=hashed, unhashed=unhashed):
                    yield f'{k}.{a}', b
            else:
                if v is None:
                    yield k, v
                elif hashed and v.is_hashed:
                    yield k, v
                elif unhashed and not v.is_hashed:
                    yield k, v


class BitstreamConfig(_FLConfigMixin):
    _settings = ('bitstream', 'ignore_version')

    def __init__(self, bitstream: str | None = None, ignore_version: bool | None = None, _hashed=None):
        """
        Args:
            bitstream: bitstream file name (full path)
            ignore_version: suppress errors related to differences in IP release verison and what driver expects
            _hashed:
        """
        self._hashed = _hashed
        if self._hashed:
            return
        self.bitstream = bitstream
        self.ignore_version = ignore_version


class RFDCClockingConfig(_FLConfigMixin):
    _settings = ('programming_key', 'clock_source')

    def __init__(self, programming_key: str | None = None, clock_source: str | None = None, _hashed=None):
        """
        Args:
            programming_key: RFDC LMX / LMK programming files
            clock_source: internal or external 10 MHz source #TODO: this should probably be made a bool
            _hashed:
        """
        self._hashed = _hashed
        if self._hashed:
            return
        self.programming_key = programming_key
        self.clock_source = clock_source


class RFDCConfig(_FLConfigMixin):
    _settings = ('dac_mts', 'adc_mts', 'adc_gains', 'dac_gains')

    def __init__(self, dac_mts: bool | None = None, adc_mts: bool | None = None, adc_gains: (float, float) = None,
                 dac_gains: (bool, bool) = None, _hashed=None):
        """
        Args:
            dac_mts: enable dac MTS
            adc_mts:  enable adc MTS (also enables DAC MTS)
            adc_gains: ADC0, ADC1. Allowed values 0-2.0
            dac_gains: DAC0, DAC1. Allowed values 0-2.0
            _hashed:
        """
        self._hashed = _hashed
        if self._hashed:
            return
        self.dac_mts = dac_mts
        self.adc_mts = adc_mts
        self.adc_gains = adc_gains
        self.dac_gains = dac_gains


class IFConfig(_FLConfigMixin):
    _settings = ('lo', 'adc_attn', 'dac_attn')

    def __init__(self, lo=None, adc_attn=None, dac_attn=None, _hashed=None):
        """
        Args:
            lo: lo position in MHz
            adc_attn: output attenuation in dB. Max is 40.
            dac_attn: input attenuation in dB. Max is 40.
            _hashed:
        """
        self._hashed = _hashed
        if self._hashed:
            return

        self.lo = lo
        self.adc_attn = adc_attn
        self.dac_attn = dac_attn


class TriggerConfig(_FLConfigMixin):
    _settings = ('holdoffs', 'thresholds')

    def __init__(self, holdoffs: np.ndarray | list = None, thresholds: np.ndarray | list = None, _hashed=None):
        """ See drivers.PhotonTrigger for notes on what these values should be"""
        self._hashed = _hashed
        if self._hashed:
            return

        self.holdoffs = np.asarray(holdoffs) if holdoffs is not None else None
        self.thresholds = np.asarray(thresholds) if thresholds is not None else None


class ChannelConfig(_FLConfigMixin):
    _settings = ('bins',)

    def __init__(self, freqs=None, bins=None, _hashed=None):
        self._hashed = _hashed
        if self._hashed:
            return
        if freqs is not None:
            bins = convert_freq_to_ddc_bins(freqs)
        self.bins = bins


class DDCConfig(_FLConfigMixin):
    _settings = ('tones', 'loop_center', 'phase_offset', 'center_relative', 'quantize')

    def __init__(self, tones=None, loop_center=None, phase_offset=None, center_relative=None,
                 quantize=None, _hashed=None):
        self._hashed = _hashed
        if self._hashed:
            return

        self.tones = tones
        self.loop_center = loop_center
        self.phase_offset = phase_offset
        self.center_relative = center_relative
        self.quantize = quantize


class FilterConfig(_FLConfigMixin):
    _settings = ('coefficients',)

    def __init__(self, coefficients=None, _hashed=None):
        self._hashed = _hashed
        if self._hashed:
            return

        self.coefficients = coefficients


class WaveformConfig(_FLConfigMixin):
    _settings = ('waveform',)

    def __init__(self, _hashed=None, waveform=None):
        """

        Args:
            qmc_settings:
            _hashed:
            waveform_values:
            waveform_spec:
            if waveform values and waveform_spec are both passed, they should be consistent
            waveform values will take precedence.
        """
        self._hashed = _hashed
        if self._hashed:
            return
        # TODO [optional] make hash use waveform_spec instead of output_waveform so that deferred computation isn't
        #  triggered by a request for the hash
        self.waveform = waveform

    @property
    def default_channel_config(self) -> ChannelConfig:
        """A convenience method to get a ChannelConfig using all the waveform's frequencies.
        Default channel config is not available for tabulated waveforms."""
        freqs = self.waveform.freqs
        bins = convert_freq_to_ddc_bins(freqs)
        return ChannelConfig(bins=bins, freqs=freqs)

    @property
    def default_ddc_config(self) -> DDCConfig:
        """A convenience method to get a DDCConfig using all the waveform's frequencies.
        Default channel config is not available for tabulated waveforms."""
        ddc_tones = np.zeros(2048)
        ddc_tones[:self.waveform.freqs.size] = self.waveform.freqs
        centers = np.zeros(2048, dtype=np.complex64)
        phase_offsets = np.zeros(2048)
        return DDCConfig(tones=ddc_tones, phase_offset=phase_offsets, loop_center=centers)


class FeedlineConfig(_FLMetaconfigMixin):
    @staticmethod
    def empty_config():
        return FeedlineConfig(bitstream=BitstreamConfig(), rfdc_clk=RFDCClockingConfig(),
                              rfdc=RFDCConfig(), if_board=IFConfig(),
                              waveform=WaveformConfig(), chan=ChannelConfig(),
                              ddc=DDCConfig(), filter=FilterConfig(),
                              trig=TriggerConfig())

    def __init__(self, bitstream: (BitstreamConfig, dict) = None,
                 rfdc_clk: (RFDCClockingConfig, dict) = None,
                 rfdc: (RFDCConfig, dict) = None,
                 if_board: (IFConfig, dict) = None,
                 waveform: (WaveformConfig, dict) = None,
                 chan: (ChannelConfig, dict) = None,
                 ddc: (DDCConfig, dict) = None,
                 filter: (FilterConfig, dict) = None,
                 trig: (TriggerConfig, dict) = None):
        self.bitstream = BitstreamConfig(**bitstream) if isinstance(bitstream, dict) else bitstream
        self.rfdc_clk = RFDCClockingConfig(**rfdc_clk) if isinstance(rfdc_clk, dict) else rfdc_clk
        self.rfdc = RFDCConfig(**rfdc) if isinstance(rfdc, dict) else rfdc
        self.if_board = IFConfig(**if_board) if isinstance(if_board, dict) else if_board
        self.waveform = WaveformConfig(**waveform) if isinstance(waveform, dict) else waveform
        self.chan = ChannelConfig(**chan) if isinstance(chan, dict) else chan
        self.ddc = DDCConfig(**ddc) if isinstance(ddc, dict) else ddc
        self.filter = FilterConfig(**filter) if isinstance(filter, dict) else filter
        self.trig = TriggerConfig(**trig) if isinstance(trig, dict) else trig

        assert isinstance(self.bitstream, (type(None), BitstreamConfig))
        assert isinstance(self.rfdc_clk, (type(None), RFDCClockingConfig))
        assert isinstance(self.rfdc, (type(None), RFDCConfig))
        assert isinstance(self.if_board, (type(None), IFConfig))
        assert isinstance(self.waveform, (type(None), WaveformConfig))
        assert isinstance(self.chan, (type(None), ChannelConfig))
        assert isinstance(self.ddc, (type(None), DDCConfig))
        assert isinstance(self.filter, (type(None), FilterConfig))
        assert isinstance(self.trig, (type(None), TriggerConfig))
        # TODO to support captures of less than all groups we need to add a self.capture_setup which has group
        # settings for iq and phase

    def __str__(self):
        def nonestr(x, name):
            return f'{name}: None' if x is None else f'{x}'.replace('\n','\n  ')

        return (f"FeedlineConfig {hash(self)}:\n"
                f"  {nonestr(self.bitstream, 'BitstreamConfig')}\n"
                f"  {nonestr(self.rfdc_clk, 'RFDCClockingConfig')}\n"
                f"  {nonestr(self.rfdc, 'RFDCConfig')}\n"
                f"  {nonestr(self.if_board, 'IFConfig')}\n"
                f"  {nonestr(self.waveform, 'WaveformConfig')}\n"
                f"  {nonestr(self.chan, 'ChannelConfig')}\n"
                f"  {nonestr(self.ddc, 'DDCConfig')}\n"
                f"  {nonestr(self.filter, 'FilterConfig')}\n"
                f"  {nonestr(self.trig, 'TriggerConfig')}")


class FeedlineConfigManager:
    def __init__(self):
        self._config = {}
        self._cache = {}
        self._current = FeedlineConfig.empty_config()

    def learn(self, config: FeedlineConfig):
        """Commit configuration info to memory for later use, hashed configurations do not add to knowledge"""
        self._cache.update({hash(v): v for _, v in config.iter(hashed=False)})  # if hash(v) not in self._cache})

    def unlearned_hashes(self, config: FeedlineConfig) -> set:
        """
        Return a set of any hashed config hashes in the config that have not been learned.

        The presence of unlearned hashes will prevent a config from being added to the manager.
        """
        return set(hash(v) for k, v in config.iter(unhashed=False) if hash(v) not in self._cache)

    def required(self) -> FeedlineConfig:
        """
        Return the feedline configuration built from all FeedlineConfigs in the manager. The config will be
        free of hashed settings but may not specify a requirement for many settings (i.e. they may be None)

        """
        # This is a new, much cleaner approach
        cfg = FeedlineConfig.empty_config()
        for config in self._config.values():
            # This adopts values specified in config (i.e. Nones are skipped). While it later configs set values are
            # overwritten
            # there is a potential concurrency issue here if the configs are being mutated, but we don't do that!!!
            cfg.merge_with(config)
        return copy.deepcopy(cfg)  # ensure the result doesn't share data!

    def pop(self, id) -> bool:
        """
        Remove a config from the manager

        Args:
            id: settings id to remove, does nothing if it is unknown

        Returns: True iff the required settings after the pop are less restrictive
        """
        if id not in self._config:
            return False
        x = self.required()
        self._config.pop(id)
        return self.required() < x

    def add(self, id, config: FeedlineConfig) -> FeedlineConfig:
        """
        Add a feedline config to the managed set of settings with an id for later removal. Configs with hashed
        settings can not be added unless their hashes have been previously learned. All unhashed settings are learned
        upon addition.

        If the config to be added has settings that are not compatible with existing settings it an error will be
        raised.

        Args:
            id: an id for the settings (for later removal from the set)
            config: a feedline config at any level of specificity

        Returns: A feedline config with the settings that need updating populated and the rest None

        Raises: ValueError if a config contains unlearned hashes.

        """
        if id in self._config and self._config[id] != config:
            raise ValueError(f'ID {id} already used by a previously added config')

        self.learn(config)
        if self.unlearned_hashes(config):
            raise ValueError('Config contains unlearned hashes')

        required = self.required()

        if not required.compatible_with(config):
            raise ValueError('Proposed settings not compatible with required settings')

        self._config[id] = config.unhashed_form(self._cache)
        new = self.required()

        # NB >= is read as "at more or equally restrictive" on the needed FPGA settings
        # so IFConfig(dac_atten=3) and IFConfig(dac_atten=4) are equally restrictive but not equal, not compatible
        # hence the need for the compatible_with check above. IFConfig(dac_atten=3) is more restrictive than
        # IFConfig(dac_atten=None) and thus no settings changes would be needed.

        for k, v in new:
            cv = getattr(self._current, k)
            if v is None or cv is None:
                continue  # our work is done or we must set for sure, either way continue

            setattr(new, k, cv.deltafy(v))  # replace with what's actually changed

        old = self._current
        self._current.merge_with(new)
        return new



import numpy as np
import pynq
from logging import getLogger
from mkidgen3.mkidpynq import check_description_for
from mkidgen3.fixedpoint import fp_factory
from mkidgen3.util import pack16_to_32
import time


class PhasematchDriver(pynq.DefaultHierarchy):
    N_TEMPLATE_TAPS = 30
    N_RES = 2048
    N_RES_P_LANE = 512
    N_LANES = 4
    N_SLOTS = 2
    N_FIFO_SIZE = 512
    COEFF_FORMAT = (1, 15)

    def __init__(self, description):
        super().__init__(description)
        self.fifo = self.reload.axi_fifo_mm_s_0
        self._pending = np.zeros(self.N_LANES, dtype=int)

    @staticmethod
    def checkhierarchy(description):
        if 'reload' not in description.get('hierarchies', {}):
            return False
        return bool(len(check_description_for(description['hierarchies']['reload'], 'xilinx.com:ip:axi_fifo_mm_s')))

    @staticmethod
    def vet_coeffs(coeffs):
        if coeffs.shape[-1] != PhasematchDriver.N_TEMPLATE_TAPS:
            raise ValueError('Incorrect number of taps')
        if np.asarray(coeffs).dtype != np.int16 and abs(np.asarray(coeffs)).max() > 1:
            raise ValueError(f'Coefficients must be <= 1 if floating point')

    @staticmethod
    def vet_res_id(res_id):
        if 0 > res_id or res_id >= PhasematchDriver.N_RES:
            raise ValueError(f'resID must be in [0-{PhasematchDriver.N_RES}-1]')

    @staticmethod
    def reorder_coeffs(coeffs):
        """convert taps to order needed by a reload packet"""
        return coeffs[::-1]  # see coefficient reload tab for order in block design

    def load_coeff(self, res_id, coeffs, vet=True, force_commit=False, raw=False, wait=False, defer_commit=False):
        """
        A reload packet consists of the coefficients and the coefficient set number

        If raw coeffs will be converted to np.uint16 via numpy casting/type coercion rules.

        See block diagram for layout. Resonators assigned to lanes 0-3 in consecutive sets of 512.

        FIRs have two reload slots and are in "on vector" update mode.

        set defer_commit to skip sending a config packet even if the packet sent filled up the number of usable slots

        See pg149 pg 18
        """
        if vet:
            self.vet_res_id(res_id)
            self.vet_coeffs(coeffs)

        if raw:
            fp_format = lambda x: x
        else:
            fp_format = fp_factory(*self.COEFF_FORMAT, True, include_index=True)

        lane = res_id % self.N_LANES
        reload_packet = np.zeros(self.N_TEMPLATE_TAPS + 1, dtype=np.uint16)
        reload_packet[0] = res_id // self.N_LANES
        reload_packet[1:] = [fp_format(c) for c in self.reorder_coeffs(coeffs)]

        cfg_packet = pack16_to_32(np.arange(self.N_RES_P_LANE, dtype=np.uint16))

        if self._pending[lane] >= self.N_SLOTS:
            getLogger(__name__).debug(f'Reload slots for lane {lane} are full, sending config packet first')
            self.fifo.tx(cfg_packet, destination=4, wait=wait)  # Send a config packet to trigger reload
            self._pending[:] = 0

        self.fifo.tx(pack16_to_32(reload_packet), destination=lane, last_bytes=2, wait=wait)  # reload channels are 0,2,4,6
        self._pending[lane] += 1

        if force_commit or (self._pending[lane] >= self.N_SLOTS and not defer_commit):
            if not force_commit:
                getLogger(__name__).debug('Sending config packet')
            self.fifo.tx(cfg_packet, destination=4, wait=wait)  # Send a config packet to trigger reload
            self._pending[:] = 0

    def load_coeff_sets(self, coeff_sets, raw=False):
        """
        Program coefficients for all the resonator channels
        Args:
            coeff_sets: (N_RES, N_TAP) array of coefficients, will be vetted
            raw: (optional) whether to load the coefficients as is or convert to fixed point, see load_coeff

        Returns: None
        """
        self.vet_coeffs(coeff_sets)
        for res in range(self.N_RES):
            self.load_coeff(res, coeff_sets[res], vet=False, defer_commit=True, force_commit=res == self.N_RES-1,
                            wait=False, raw=raw)

    def configure(self, coefficients=None):
        if coefficients is None:
            return
        getLogger(__name__).info(f'Configuring phasematch with {coefficients}')
        if isinstance(coefficients, str) and coefficients.startswith('unity'):
            try:
                n = min(max(1, int(coefficients.strip('unity'))), 2048)
            except:
                n = 2048
            coefficients = np.zeros((2048, self.N_TEMPLATE_TAPS), dtype=np.int16)
            coefficients[:n, 0] = 2 ** 15 - 1
        if coefficients.shape != (2048, self.N_TEMPLATE_TAPS) or coefficients.dtype != 'int16':
            raise ValueError(f'coefficients must be a ({self.N_RES},{self.N_TEMPLATE_TAPS}) int16 array')

        self.load_coeff_sets(coefficients, raw=True)



import zmq
import logging
logging.basicConfig(level=logging.DEBUG)
from logging import getLogger
from mkidgen3.server.feedline_config import WaveformConfig
from mkidgen3.server.captures import CaptureRequest, CaptureJob, FeedlineConfig


"""
Setting up from scratch
1. Run sweeps (power and freq.) to find res and drive power
2. Process
3. Rerun 1&2 with fixed freq to finialize optimal drive power
4. Run IQ sweeps to find loop centers
5. Process
6. capture Optimal filter phase data
7. Process
8. capture phase data for thresholding
9. Process
10. ready to observe

Observing

IrrOps
Observing, feedline is acting weird, what do?
0. stop observing or cut out feedline
1. e.g. reset/power cycle/replace feedline
2. resume


"""



ctx = zmq.Context.instance()
ctx.linger = 0

# cap command default 8888
# cap data 8889
# cap status 9000

feedline_server = 'tcp://localhost:8888'
capture_data_server = 'tcp://localhost:8889'
status_server = 'tcp://localhost:8890'


frequencies = []
coefficients = []
thresholds= []
holdoffs = []
fc = FeedlineConfig(bitstream=dict(bitstream=None, ignore_version=None),
                    rfdc_clk=dict(programming_key=None, clock_source=None),
                    rfdc=dict(dac_mts=True, adc_mts=False, adc_gains=None, dac_gains=None),
                    if_board=dict(lo=None, adc_atten=None, dac_atten=None),
                    waveform=dict(frequencies=None, amplitudes=None, phases=None, iq_ratios=None, phase_offsets=None, maximize_dynamic_range=True),
                    chan=dict(frequencies=frequencies),
                    ddc=dict(tones=None, loop_center=None, phase_offset=None, center_relative=None, quantize=None),
                    filter=dict(coefficients=coefficients),
                    trig=dict(thresholds=thresholds, holdoffs=holdoffs))
cr = CaptureRequest(1337, 'iq', fc)
cj = CaptureJob(cr, feedline_server, capture_data_server, status_server, submit=False)
cj.submit()
x = cj.data(timeout=5)  #should be almost
print(cj.status_history())

#Start the capture of photons from an MKID array, assume FeedlineReadoutServers are running on all the boards
feedline_server_urls = []
data_server_urls = []
status_server_urls = []

#Generate a suitable FeedlineConfig for each server
fc = FeedlineConfig()
buff_duration_ms = 100
jobs = []
for a,b,c in zip(feedline_server_urls, data_server_urls, status_server_urls):
    cr = CaptureRequest(buff_duration_ms, 'photons', fc, feedline_server=a)
    jobs.append(CaptureJob(cr, a, b, c, submit=True))




cap_data_urls = []
from zmq.devices import ThreadDevice

data_server_internal = 'inproc://cap_data.xpub'
# Set up a proxy for routing all the capture requests
dtd = ThreadDevice(zmq.QUEUE, zmq.XSUB, zmq.XPUB)
dtd.setsockopt_in(zmq.LINGER, 0)
dtd.setsockopt_out(zmq.LINGER, 0)
for url in cap_data_urls:
    dtd.connect_in(url)
dtd.bind_out(data_server_internal)
dtd.daemon = True
dtd.start()
getLogger(__name__).info(f'Relaying all capture data from {cap_data_urls} to {data_server_internal}')



import logging
from logging import getLogger
import numpy as np
import time
import pynq
from mkidgen3.fixedpoint import FP16_15


class DACTableURAM(pynq.DefaultHierarchy):
    @staticmethod
    def checkhierarchy(description):
        # found = check_description_for(description, ('xilinx.com:axis_switch', 'mazinlab:mkidgen3:filter_iq',
        #                                             'xilinx.com:module_ref:axis2mm'))
        return description['fullpath'] == 'dactable'  # and bool(len(found['xilinx.com:module_ref:axis2mm']))

    def __init__(self, description):
        super().__init__(description)
        self._buffer = None

    def replay_ramp(self):
        ramp = np.arange(2 ** 19, dtype=np.complex64)
        ramp.real %= (2 ** 16)
        ramp.imag = ramp.real
        self.replay(ramp, fpgen=None)

    def replay(self, data, fpgen=lambda x: [FP16_15(v).__index__() for v in x]):
        """
        data - an array of complex numbers, nominally 2^19 samples
        fpgen - set to a function to convert floating point numbers to integers. must work on arrays of data
        if None data will be truncated.
        """
        if fpgen == 'simple':
            data = (data * 8192).round().clip(-8192, 8191) * 4
            fpgen = None
        # Data has right shape
        if data.size < 2 ** 19:
            getLogger(__name__).warning('Insufficient data, padding with zeros')
            x = np.zeros(2 ** 19, dtype=np.complex64)
            x[:data.size] = data
            data = x

        if data.size > 2 ** 19:
            getLogger(__name__).warning('Data will be truncated to 2^19 samples')
            data = data[:2 ** 19]

        self._buffer = np.zeros(2 ** 20, dtype=np.uint16)
        iload = fpgen(data.real) if fpgen is not None else data.real.astype(np.int16)
        qload = fpgen(data.imag) if fpgen is not None else data.imag.astype(np.int16)
        for i in range(8):
            self._buffer[i:data.size * 2:16] = iload[i::8]
            self._buffer[i + 8:data.size * 2:16] = qload[i::8]
        self.axi_bram_ctrl_0.mmio.array[:] = np.frombuffer(self._buffer, dtype=np.uint32)
        time.sleep(1)

    def stop(self):
        self._buffer = np.zeros(2 ** 20, dtype=np.uint16)
        self.axi_bram_ctrl_0.mmio.array[:] = 0

    def quiet(self):
        """Replay 0s"""
        self.stop()

    def status(self)->dict:
        """
        Generate a dictionary with whether the core is running and a copy of the current buffer
        Returns: dict with keys running and buffer
        """
        return {'running': True, 'buffer': self._buffer.copy() if self._buffer is not None else None}

    def configure(self, waveform=None):
        """
        Args:
            waveform_values: complex values quantized to integer real and imaginary parts
            max val, min val: [2**15-1, -2**15] corresponding to [1.0, -1.0] & [max dac V, min dac V]
        Returns: None
        """
        try:
            waveform_values = waveform.output_waveform
        except AttributeError:
            getLogger(__name__).info('Interpreting waveform as array')
            waveform_values = waveform

        if waveform_values is None:
            raise ValueError('waveform_values is None')

        self.replay(waveform_values, fpgen=None)


class DACTableAXIM(pynq.DefaultIP):
    bindto = ['mazinlab:mkidgen3:dac_table_axim:1.33']

    def __init__(self, description):
        super().__init__(description=description)
        self._buffer = None

    def replay_ramp(self, tlast_every=256):
        ramp = np.arange(2 ** 19, dtype=np.complex64)
        ramp.real %= (2 ** 16)
        ramp.imag = ramp.real
        self.stop()
        self.replay(ramp, tlast_every=tlast_every, fpgen=None)

    def replay(self, data, tlast=True, tlast_every=256, replay_len=None, start=True,
               fpgen=lambda x: [FP16_15(v).__index__() for v in x], stop_if_needed=False):
        """
        data - an array of complex numbers, nominally 2^19 samples
        tlast - Set to true to emit a tlast pulse every tlast_every transactions
        tlast_every - assert tlast every tlast_every*16 th sample.
        replay_len - Number of groups of 16 (DAC takes 16/clock) in the data to use.
            e.g. if data[N].reshape(N//16,16). Note that the user channel will count to length-1
        start - start the replay immediately
        fpgen - set to a function to convert floating point numbers to integers. must work on arrays of data
        if None data will be truncated.
        """
        if self.register_map.run.run and not stop_if_needed:
            raise RuntimeError('Replay in progress. Call .stop() or with stop_if_needed=True')
        if fpgen == 'simple':
            data = (data * 8192).round().clip(-8192, 8191) * 4
            fpgen = None
        # Data has right shape
        if data.size < 2 ** 19:
            getLogger(__name__).warning('Insufficient data, padding with zeros')
            x = np.zeros(2 ** 19, dtype=np.complex64)
            x[:data.size] = data
            data = x

        if data.size > 2 ** 19:
            getLogger(__name__).warning('Data will be truncated to 2^19 samples')
            data = data[:2 ** 19]

        # Process Length
        if replay_len is None:
            replay_len = data.size // 16
        replay_len = int(replay_len)

        if replay_len > 2 ** 15:  # NB 15 as length is in sets of 16
            getLogger(__name__).warning('Clipping replay length to 2^15 sets of 16')
            replay_len = 2 ** 15
        if replay_len % 2:
            raise ValueError('Replay length must be a multiple of 2')

        # Process tlast_every
        tlast_every = int(tlast_every)

        if tlast and not (1 <= tlast_every <= replay_len):
            raise ValueError('tlast_every must be in [1-length] when tlast is set')
        if not tlast:
            tlast_every = 256  # just assign some value to ensure we didn't get handed garbage

        if self._buffer is not None:
            self._buffer.freebuffer()
        self._buffer = pynq.allocate(2 ** 20, dtype=np.uint16)

        iload = fpgen(data.real) if fpgen is not None else data.real.astype(np.int16)
        qload = fpgen(data.imag) if fpgen is not None else data.imag.astype(np.int16)
        for i in range(16):
            self._buffer[i:data.size * 2:32] = iload[i::16]
            self._buffer[i + 16:data.size * 2:32] = qload[i::16]

        if self.register_map.run.run:
            self.stop()
            while not self.register_map.CTRL.AP_IDLE:
                time.sleep(.0001)
        self.register_map.a_1 = self._buffer.device_address
        self.register_map.length_r = replay_len - 1  # length counter
        self.register_map.tlast = bool(tlast)
        self.register_map.replay_length = tlast_every - 1
        self.register_map.run = True
        if start:
            self.register_map.CTRL.AP_START = 1

    def stop(self):
        # Note the dac still outputs stuff (harmonics of last buffer)
        # Need to load URAM with zeros if you want it to be quiet (see quiet)
        self.register_map.run = False

    def quiet(self):
        """Replay 0s"""
        self.replay(np.zeros(16, dtype=np.complex64), tlast=False, replay_len=16, stop_if_needed=True)
        self.stop()

    def status(self)->dict:
        """
        Generate a dictionary with whether the core is running and a copy of the current buffer
        Returns: dict with keys running and buffer
        """
        return {'running': self.register_map.run.run and self.register_map.CTRL.AP_START,
                'buffer': self._buffer.copy() if self._buffer is not None else None}

    def configure(self, waveform=None):
        """
        Args:
            waveform_values: complex values quantized to integer real and imaginary parts
            max val, min val: [2**15-1, -2**15] corresponding to [1.0, -1.0] & [max dac V, min dac V]
        Returns: None
        """
        try:
            waveform_values = waveform.output_waveform
        except AttributeError:
            getLogger(__name__).info('Interpreting waveform as array')
            waveform_values = waveform

        if waveform_values is None:
            raise ValueError('waveform_values is None')

        self.replay(waveform_values, tlast=True, tlast_every=256, replay_len=None, start=True,
                    fpgen=None, stop_if_needed=True)



import time
import pickle

import numpy as np

from dataclasses import dataclass, field
from typing import Optional, Any


from mkidgen3.server.waveform import SimpleFreqlistWaveform
from mkidgen3.server.feedline_config import WaveformConfig
from mkidgen3.mkid_setup.sweeps import (
    SweepConfig,
    PowerSweepConfig,
    PowerSweep,
    Sweep,
)


def find_maxiq(freqs, iqs, start):
    i = start
    while np.abs(iqs[i - 1] - iqs[i]) > np.abs(iqs[i + 1] - iqs[i]) or np.abs(
        iqs[i - 1]
    ) < np.abs(iqs[i]):
        i -= 1
    return freqs[i]


def agc(ifb, capture, output_atten, target_fraction=0.2, step=1, input_atten=62):
    while input_atten > 0:
        ifb.set_attens(output_attens=output_atten, input_attens=input_atten)
        d = capture.capture_adc(2**19, complex=False)
        m = np.max(np.abs(d))
        del d
        if m / 32764.0 > target_fraction:
            ifb.set_attens(output_attens=output_atten, input_attens=input_atten + step)
            return input_atten + step
        input_atten -= step
    return 0.0


def fft_agc(overlay, target_fraction=0.2, caplen=4096):
    options = [
        0xFFF,
        0xF7F,
        0x77F,
        0x777,
        0x757,
        0x755,
        0x555,
        0x515,
        0x115,
        0x111,
        0x101,
        0x001,
        0x000,
    ]
    gpio = overlay.photon_pipe.opfb.fft.axi_gpio_0
    for i in range(1, len(options)):
        gpio.channel1.write(options[i], 0xFFF)
        light = overlay.capture.capture_iq(caplen, tap_location="ddciq")
        blight = np.copy(light).flatten()
        del light
        if np.max(np.abs(blight)) / (1 << 15) > target_fraction:
            gpio.channel1.write(options[i - 1], 0xFFF)
            return options[i - 1]
    gpio.channel1.write(options[-1], 0xFFF)
    return options[-1]


@dataclass
class BinConfig:
    fftbin: np.int16
    tone: np.float64
    phase_offset: np.float64
    loop_center: np.complex64


@dataclass
class IQCapture:
    filename: str
    savetime: float
    captime: tuple[int, int, int]
    channel_map: Optional[dict[int, int]]
    usermeta: Any

    def load_channel(self, base, channel, sl=None):
        if self.channel_map:
            channel = self.channel_map[channel]
        f = np.load(base + self.filename)
        if sl:
            return f[sl, channel, 0] + 1.0j * f[sl, channel, 1]
        return f[::, channel, 0] + 1.0j * f[::, channel, 1]


@dataclass
class MKIDSetup:
    name: str
    attenuation: tuple[float, float]
    waveform: SimpleFreqlistWaveform
    bin_settings: list[list[BinConfig]]
    darks: list[list[IQCapture]] = field(default_factory=list)
    lights: list[dict[float | str, list[IQCapture]]] = field(default_factory=list)
    powersweep: Optional[PowerSweep] = None
    biassweep: Optional[Sweep] = None
    centeringsweeps: list[Sweep] = field(default_factory=list)
    centeredsweeps: list[Sweep] = field(default_factory=list)
    lo: float = 6000.0
    fftscale: int = 0xFFF
    tied_tones: dict[int, list[int]] = field(default_factory=dict)
    basedir: str = "./"

    def load(self, overlay, ifboard, *, recenter: bool = False):
        ol, ifb = overlay, ifboard

        ifb.set_lo(self.lo)
        ifb.set_attens(*self.attenuation)
        gpio = overlay.photon_pipe.opfb.fft.axi_gpio_0
        gpio.channel1.write(self.fftscale, 0xFFF)

        waveform = WaveformConfig(waveform=self.waveform)
        if self.bin_settings and not recenter:
            bins = np.array([b.fftbin for b in self.bin_settings[-1]])
            btones = np.array([b.tone for b in self.bin_settings[-1]])
            bphase = np.array([b.phase_offset for b in self.bin_settings[-1]])
            bcent = np.array([b.loop_center for b in self.bin_settings[-1]])
        else:
            bins = waveform.default_channel_config.bins
            btones = waveform.default_ddc_config.settings_dict()["tones"]
            bphase = np.zeros_like(btones)
            bcent = np.zeros_like(btones, dtype=np.complex64)

        print(btones)

        ol.dac_table.configure(**waveform.settings_dict())
        ol.photon_pipe.reschan.bin_to_res.configure(**{"bins": bins})
        ol.photon_pipe.reschan.ddccontrol_0.configure(
            **{"tones": btones, "phase_offset": -bphase, "loop_center": bcent}
        )

        if recenter:
            from mkidgen3.mkid_setup.loop_locator import rotate_and_center

            print("Centering...")
            centeringsweep = SweepConfig.from_bandwidth(
                bandwidth=0.25e6,
                points=256,
                average=1024 * 64,
                waveform=waveform,
                lo_center=self.lo,
                attens=self.attenuation,
            ).run_sweep(ifboard=ifb, capture=ol.capture)

            config, _, _ = rotate_and_center(
                centeringsweep, targets=list(np.zeros(len(self.waveform.quant_freqs)))
            )
            bphase = config.settings_dict()["phase_offset"][: len(bphase)]
            bcent = config.settings_dict()["loop_center"][: len(bphase)]
            ol.photon_pipe.reschan.ddccontrol_0.configure(
                **{"tones": btones, "phase_offset": -bphase, "loop_center": bcent}
            )

            centeredsweep = SweepConfig.from_bandwidth(
                bandwidth=0.25e6,
                points=256,
                average=1024 * 64,
                waveform=waveform,
                lo_center=self.lo,
                attens=self.attenuation,
            ).run_sweep(ifboard=ifb, capture=ol.capture)

            bin_setting = [
                BinConfig(bins[i], btones[i], bphase[i], bcent[i])
                for i in range(len(bins))
            ]

            self.bin_settings.append(bin_setting)
            self.centeringsweeps.append(centeringsweep)
            self.centeredsweeps.append(centeredsweep)
            self.darks.append([])
            self.lights.append({})

    def take_snap(
        self, name, overlay, first_sixteen=False, metacallback=lambda: None, groups=None
    ):
        #           Mib  Kib  b
        memsize = 4 * 1024 * 1024 * 1024
        sample_size = 16 * 2 * 2 if first_sixteen or groups else 16 * 2 * 2048
        count = memsize // (sample_size * 2)

        if first_sixteen:
            groups = [0, 1]

        data = overlay.capture.capture_iq(count, groups=groups, tap_location="ddciq")
        captime = (
            overlay.capture.axis2mm.mmio.read(0b01 << 2),
            overlay.capture.axis2mm.mmio.read(0b10 << 2),
            overlay.capture.axis2mm.mmio.read(0b11 << 2),
        )
        savetime = time.time()
        fn = "{:s}-{:s}-{:f}.npy".format(self.name, name, savetime)
        if first_sixteen:
            np.save(self.basedir + fn, data[::, :16, ::])
        else:
            np.save(self.basedir + fn, data[::, : len(self.waveform.quant_freqs), ::])
        data.freebuffer()
        del data

        channel_map = None
        if groups:
            channel_map = {}
            for i in range(16):
                channel_map[groups[0] * 16 + i] = i

        return IQCapture(fn, savetime, captime, channel_map, metacallback()), count / 2e6

    def take_darks(
        self,
        overlay,
        *,
        count=None,
        time=None,
        first_sixteen=False,
        metacallback=lambda: None,
        groups=None,
    ):
        import tqdm.autonotebook as tqdm

        assert (count is None) or (time is None)
        assert (count is not None) or (time is not None)
        if count:
            for i in tqdm.tqdm(range(count), desc="Darks"):
                sn, t = self.take_snap(
                    "dark",
                    overlay,
                    first_sixteen=first_sixteen,
                    metacallback=metacallback,
                    groups=groups,
                )
                self.darks[-1].append(sn)
        else:
            total = 0
            with tqdm.tqdm(total=time, desc="Darks") as pbar:
                while total < time:
                    sn, t = self.take_snap(
                        "dark",
                        overlay,
                        first_sixteen=first_sixteen,
                        groups = groups,
                        metacallback=metacallback,
                    )
                    self.darks[-1].append(sn)
                    total += t
                    pbar.update(t)

    def take_lights(
        self,
        overlay,
        source,
        *,
        count=None,
        time=None,
        first_sixteen=False,
        metacallback=lambda: None,
        groups=None,
    ):
        import tqdm.autonotebook as tqdm

        assert (count is None) or (time is None)
        assert (count is not None) or (time is not None)
        if source not in self.lights[-1].keys():
            self.lights[-1][source] = []
        if count:
            for i in tqdm.tqdm(range(count), desc="Lights"):
                sn, t = self.take_snap(
                    "light",
                    overlay,
                    first_sixteen=first_sixteen,
                    metacallback=metacallback,
                    groups=groups,
                )
                self.lights[-1][source].append(sn)
        else:
            total = 0
            with tqdm.tqdm(total=time, desc="Lights") as pbar:
                while total < time:
                    sn, t = self.take_snap(
                        "light",
                        overlay,
                        first_sixteen=first_sixteen,
                        metacallback=metacallback,
                        groups=groups,
                    )
                    self.lights[-1][source].append(sn)
                    total += t
                    pbar.update(t)

    def save(self, *, savetime=True):
        pickle.dump(
            self,
            open(
                self.basedir
                + self.name
                + (".pkl" if not savetime else "{:f}.pkl".format(time.time())),
                "wb",
            ),
            protocol=pickle.HIGHEST_PROTOCOL,
        )
    @classmethod
    def with_selected(cls, name, overlay, ifboard, starting_tones, starting_amps, output_atten, selected_attens, *, maxvel=False, lo=6000.0, target_dynrange=0.25, tied_tones={}):
        import tqdm.autonotebook as tqdm

        from mkidgen3.mkid_setup.loop_locator import rotate_and_center

        baseline = output_atten

        ol, ifb = overlay, ifboard
        spacing = np.fft.fftfreq(1 << 19, 1 / 4.096e9)[1]

        tones = np.array(starting_tones)
        amps = np.array(starting_amps)

        attens = np.array(selected_attens)
        amps = amps * (10 ** (-(attens - baseline) / 20))

        print("Configuring DAC...")

        waveform = WaveformConfig(
            waveform=SimpleFreqlistWaveform(frequencies=tones, amplitudes=amps)
        )
        ol.dac_table.configure(**waveform.settings_dict())
        chan = waveform.default_channel_config
        ol.photon_pipe.reschan.bin_to_res.configure(**chan.settings_dict())
        ddc = waveform.default_ddc_config
        ol.photon_pipe.reschan.ddccontrol_0.configure(**ddc.settings_dict())

        print("Running AGC...")
        ifb.set_lo(lo + 5)
        atten = (baseline, agc(ifb, ol.capture, baseline, target_dynrange, 0.25))
        fftsetting = fft_agc(ol, target_dynrange)
        ifb.set_lo(lo)

        if maxvel:
            print("Finding MAX IQ Velocity...")
            sweepmaxvel = SweepConfig(
                np.arange(-128, 16) * spacing / 1e6,
                waveform,
                lo,
                attens=atten,
                average=1024 * 64,
            ).run_sweep(ifboard=ifb, capture=ol.capture)
            start = np.argmin(np.abs(sweepmaxvel.config.steps))
            newtones = np.array(
                [
                    find_maxiq(sweepmaxvel.frequencies[i, ::], sweepmaxvel.iq[i, ::], start)
                    - lo * 1e6
                    for i in range(sweepmaxvel.frequencies.shape[0])
                ]
            )

            for main, tieds in tied_tones.items():
                newtones[tieds] = tones[tieds] + (newtones[main] - tones[main])
            tones = newtones

            waveform = WaveformConfig(
                waveform=SimpleFreqlistWaveform(frequencies=tones, amplitudes=amps)
            )
            ol.dac_table.configure(**waveform.settings_dict())
            chan = waveform.default_channel_config
            ol.photon_pipe.reschan.bin_to_res.configure(**chan.settings_dict())
            ddc = waveform.default_ddc_config
            ol.photon_pipe.reschan.ddccontrol_0.configure(**ddc.settings_dict())

            print("Rerunning AGC...")
            ifb.set_lo(lo + 5)
            atten = (baseline, agc(ifb, ol.capture, baseline, target_dynrange, 0.25))
            fftsetting = fft_agc(ol, target_dynrange)
            ifb.set_lo(lo)
        else:
            sweepmaxvel = None

        setup = MKIDSetup(
            name,
            atten,
            waveform.waveform,
            [],
            powersweep=None,
            biassweep=sweepmaxvel,
            fftscale=fftsetting,
            tied_tones=tied_tones,
            lo=lo,
        )
        setup.load(ol, ifb, recenter=True)
        return setup


    @classmethod
    def interactive_setup(
        cls,
        name,
        overlay,
        ifboard,
        starting_tones,
        starting_amps,
        output_attens,
        *,
        lo=6000.0,
        target_dynrange=0.25,
        tied_tones={},
    ):
        import matplotlib.pyplot as plt
        import ipywidgets as widgets
        import tqdm.autonotebook as tqdm
        from IPython.display import display
        from jupyter_ui_poll import ui_events

        from mkidgen3.mkid_setup.loop_locator import rotate_and_center

        ol, ifb = overlay, ifboard
        spacing = np.fft.fftfreq(1 << 19, 1 / 4.096e9)[1]

        tones = np.array(starting_tones)
        amps = np.array(starting_amps)
        waveform = WaveformConfig(
            waveform=SimpleFreqlistWaveform(frequencies=tones, amplitudes=amps)
        )
        ol.dac_table.configure(**waveform.settings_dict())
        chan = waveform.default_channel_config
        ol.photon_pipe.reschan.bin_to_res.configure(**chan.settings_dict())
        ddc = waveform.default_ddc_config
        ol.photon_pipe.reschan.ddccontrol_0.configure(**ddc.settings_dict())

        fftscales = {}
        attens = {}
        ifb.set_lo(lo + 5)
        for atten in tqdm.tqdm(output_attens, desc="Running AGC..."):
            attens[atten] = agc(ifb, ol.capture, atten, target_dynrange, 0.25)
            fftscales[atten] = fft_agc(ol, target_dynrange)
        ifb.set_lo(lo)

        gpio = overlay.photon_pipe.opfb.fft.axi_gpio_0
        gpio.channel1.write(max(fftscales.values()), 0xFFF)

        sweep_base = SweepConfig(
            np.arange(-128, 64) * spacing / 1e6,
            waveform,
            lo,
            average=1024 * 16,
            rmses=False,
        )
        powersweep = PowerSweepConfig(
            attens=attens, sweep_config=sweep_base
        ).run_powersweep(ifb, ol.capture, progress=True)

        current_channel = 0
        current_atten = 0
        atten_options = np.array(list(powersweep.sweeps.keys()))
        atten_settings = np.zeros(len(tones), dtype=np.int64) - 1

        prev_res = widgets.Button(description="Prev")
        up = widgets.Button(description="Up")
        down = widgets.Button(description="Down")
        next_res = widgets.Button(description="Next")

        done = widgets.Button(description="DONE")

        cont = False

        def on_up(b):
            nonlocal current_atten
            nonlocal current_channel
            current_atten += 1
            if current_atten < 0:
                current_atten = 0
            if current_atten >= len(atten_options):
                current_atten = len(atten_options) - 1
            atten_settings[current_channel] = current_atten
            update_plot()

        def on_down(b):
            nonlocal current_atten
            nonlocal current_channel
            current_atten -= 1
            if current_atten < 0:
                current_atten = 0
            if current_atten >= len(atten_options):
                current_atten = len(atten_options) - 1
            atten_settings[current_channel] = current_atten
            update_plot()

        def on_next(b):
            nonlocal current_atten
            nonlocal current_channel
            current_channel += 1
            if current_channel < 0:
                current_channel = 0
            if current_channel >= len(tones):
                current_channel = len(tones) - 1
            current_atten = (
                atten_settings[current_channel]
                if atten_settings[current_channel] != -1
                else current_atten
            )
            atten_settings[current_channel] = current_atten
            update_plot()

        def on_prev(b):
            nonlocal current_atten
            nonlocal current_channel
            current_channel -= 1
            if current_channel < 0:
                current_channel = 0
            if current_channel >= len(tones):
                current_channel = len(tones) - 1
            current_atten = (
                atten_settings[current_channel]
                if atten_settings[current_channel] != -1
                else current_atten
            )
            atten_settings[current_channel] = current_atten
            update_plot()

        def on_done(b):
            nonlocal cont
            cont = True

        prev_res.on_click(on_prev)
        up.on_click(on_up)
        down.on_click(on_down)
        next_res.on_click(on_next)
        done.on_click(on_done)

        display(widgets.HBox([prev_res, up, down, next_res, done]))

        fig, (ax1, ax2) = plt.subplots(1, 2)

        def update_plot():
            nonlocal current_atten
            nonlocal current_channel
            ax1.clear()
            ax2.clear()
            ax1.set_title(atten_options[current_atten])
            ax2.set_title(current_channel)
            powersweep.sweeps[atten_options[current_atten]][1].plot(
                ax1,
                channels=slice(current_channel, current_channel + 1),
                newtones=list(tones),
            )
            powersweep.sweeps[atten_options[current_atten]][1].plot_loops(
                ax2,
                channels=slice(current_channel, current_channel + 1),
                newtones=list(tones),
            )
            fig.canvas.draw()

        update_plot()

        with ui_events() as poll:
            while not cont:
                poll(10)
        for main, tieds in tied_tones.items():
            atten_settings[tieds] = atten_settings[main]
        print("Selected attens:", atten_options[atten_settings])
        basesel = widgets.FloatText(
            value=atten_options[atten_settings[0]],
            step=0.25,
            description="Select Baseline Atten:",
            disabled=False,
        )
        done = widgets.Button(description="DONE")
        done.on_click(on_done)
        cont = False
        display(widgets.HBox([basesel, done]))
        with ui_events() as poll:
            while not cont:
                poll(10)
        baseline = basesel.value

        attens = atten_options[atten_settings]
        amps = amps * (10 ** (-(attens - baseline) / 20))

        print("Configuring DAC...")

        waveform = WaveformConfig(
            waveform=SimpleFreqlistWaveform(frequencies=tones, amplitudes=amps)
        )
        ol.dac_table.configure(**waveform.settings_dict())
        chan = waveform.default_channel_config
        ol.photon_pipe.reschan.bin_to_res.configure(**chan.settings_dict())
        ddc = waveform.default_ddc_config
        ol.photon_pipe.reschan.ddccontrol_0.configure(**ddc.settings_dict())

        print("Rerunning AGC...")
        ifb.set_lo(lo + 5)
        atten = (baseline, agc(ifb, ol.capture, baseline, target_dynrange, 0.25))
        fftsetting = fft_agc(ol, target_dynrange)
        ifb.set_lo(lo)

        print("Finding MAX IQ Velocity...")
        sweepmaxvel = SweepConfig(
            np.arange(-128, 16) * spacing / 1e6,
            waveform,
            lo,
            attens=atten,
            average=1024 * 64,
        ).run_sweep(ifboard=ifb, capture=ol.capture)
        start = np.argmin(np.abs(sweepmaxvel.config.steps))
        newtones = np.array(
            [
                find_maxiq(sweepmaxvel.frequencies[i, ::], sweepmaxvel.iq[i, ::], start)
                - lo * 1e6
                for i in range(sweepmaxvel.frequencies.shape[0])
            ]
        )

        for main, tieds in tied_tones.items():
            newtones[tieds] = tones[tieds] + (newtones[main] - tones[main])
        tones = newtones

        waveform = WaveformConfig(
            waveform=SimpleFreqlistWaveform(frequencies=tones, amplitudes=amps)
        )
        ol.dac_table.configure(**waveform.settings_dict())
        chan = waveform.default_channel_config
        ol.photon_pipe.reschan.bin_to_res.configure(**chan.settings_dict())
        ddc = waveform.default_ddc_config
        ol.photon_pipe.reschan.ddccontrol_0.configure(**ddc.settings_dict())

        print("Rerunning AGC...")
        ifb.set_lo(lo + 5)
        atten = (baseline, agc(ifb, ol.capture, baseline, target_dynrange, 0.25))
        fftsetting = fft_agc(ol, target_dynrange)
        ifb.set_lo(lo)

        setup = MKIDSetup(
            name,
            atten,
            waveform.waveform,
            [],
            powersweep=powersweep,
            biassweep=sweepmaxvel,
            fftscale=fftsetting,
            tied_tones=tied_tones,
            lo=lo,
        )
        setup.load(ol, ifb, recenter=True)
        return setup

    @property
    def dark(self):
        for ds in self.darks[::-1]:
            if ds:
                return ds[-1]
        return None

    @property
    def by_source(self):
        sources = {}
        for ls in self.lights:
            for k, v in ls.items():
                if k not in sources:
                    sources[k] = []
                sources[k].extend(v)
        return sources



import os
import multiprocessing

import h5py
import hdf5plugin

import numpy as np
import numpy.typing as nt

from typing import Optional

TABLE_TYPE = np.dtype(
    [("resID", "<u4"), ("time", "<u4"), ("wavelength", "<f4"), ("weight", "<f4")]
)

DEFAULT_FLAGS = (
    "beammap.noDacTone",
    "beammap.failed",
    "beammap.yFailed",
    "beammap.xFailed",
    "beammap.double",
    "beammap.wrongFeedline",
    "beammap.duplicatePixel",
    "flatcal.bad",
    "flatcal.not_all_weights_valid",
    "pixcal.hot",
    "pixcal.cold",
    "pixcal.dead",
    "wavecal.bad",
    "wavecal.failed_validation",
    "wavecal.failed_convergence",
    "wavecal.not_monotonic",
    "wavecal.not_enough_histogram_fits",
    "wavecal.not_attempted",
    "wavecal.no_histograms",
    "wavecal.histogram_fit_problems",
    "wavecal.linear",
    "wavecal.quadratic",
)


def _pyts(s: str | bytes | h5py.Empty):
    if not s:
        return h5py.Empty("S1")
    if type(s) is bytes:
        return np.array(s, dtype="S")
    if type(s) is h5py.Empty:
        return s
    return np.array(bytes(s, "utf-8"), dtype="S")


def _styp(s):
    if type(s) is h5py.Empty:
        return ""
    return str(bytes(s[::]), encoding="utf-8")


def _pytable_compat(node, name: bytes, c: bytes):
    pytc = _pyts(c)
    pytn = _pyts(name)
    if c == b"TABLE":
        pytv = _pyts(b"2.6")
    elif c == b"ARRAY":
        pytv = _pyts(b"2.4")
    elif c == b"GROUP":
        pytv = _pyts(b"1.0")
    elif c == b"FILE":
        pytc = _pyts(b"GROUP")
        pytv = _pyts(b"1.0")
    else:
        raise ValueError("Pytables class {:s} unsupported".format(c))

    node.attrs["VERSION"] = pytv
    node.attrs["CLASS"] = pytc
    node.attrs["TITLE"] = pytn
    if c == b"FILE":
        node.attrs["PYTABLES_FORMAT_VERSION"] = _pyts("2.0")
    if c == b"ARRAY":
        node.attrs["FLAVOR"] = _pyts("numpy")


class TableWriter:
    _opened = False

    def __init__(
        self,
        filename: str,
        wavelength_bounds: tuple[np.int64 | int, np.int64 | int],
        compress_kwargs=hdf5plugin.Blosc(),
    ):
        self._resize_lock = multiprocessing.Lock()
        self.f = f = h5py.File(filename, "w")
        self._opened = True
        _pytable_compat(f, b"MKID Photon File", b"FILE")
        self._photons = f.create_group("photons")
        _pytable_compat(self._photons, b"Photon Information", b"GROUP")
        self._pt = pt = self._photons.create_dataset(
            "photontable",
            shape=(0,),
            dtype=TABLE_TYPE,
            maxshape=(None,),
            chunks=True,
            **compress_kwargs
        )
        for i, n in enumerate(TABLE_TYPE.names):
            pt.attrs["FIELD_{:d}_FILL".format(i)] = TABLE_TYPE[i].type(0)
            pt.attrs["FIELD_{:d}_NAME".format(i)] = _pyts(n)
        _pytable_compat(pt, b"Photon Datatable", b"TABLE")

        pt.attrs["max_wavelength"] = max(wavelength_bounds)
        pt.attrs["min_wavelength"] = min(wavelength_bounds)

        self.with_start_time()
        self.with_flags()
        self.with_dead_time()
        self.with_energy_resolution()
        self.with_data_dir(str(os.path.dirname(os.path.abspath(filename))))

    def append_photons(self, photons: nt.NDArray[TABLE_TYPE]):
        if len(photons.shape) != 1:
            raise TypeError("Expected 1d array of photons")
        self._resize_lock.acquire()
        index = self._pt.size
        self._pt.resize((self._pt.size + photons.size,))
        self._pt.attrs["NROWS"] = np.int64(index + photons.size)
        self._resize_lock.release()

        self._pt[index : index + photons.size] = photons[::]
        return self

    def with_beammap(
        self, file: str, resids: nt.NDArray[np.int64], flags: nt.NDArray[np.int64]
    ):
        if resids.shape != flags.shape:
            raise TypeError("resids and flags must have the same shape")
        self._pt.attrs["E_BMAP"] = _pyts(str(file))
        b = self.f.create_group("beammap")
        _pytable_compat(b, b"Beammap Information", b"GROUP")
        resid = b.create_dataset("map", data=resids)
        flag = b.create_dataset("flag", data=flags)
        _pytable_compat(resid, b"resID map", b"ARRAY")
        _pytable_compat(flag, b"flag map", b"ARRAY")
        return self

    def with_start_time(self, t: Optional[np.int64 | int] = None):
        if t is None:
            import time

            t = time.time()
        t = np.int64(t)
        self._pt.attrs["UNIXSTR"] = t
        return self

    def with_end_time(self, t: Optional[np.int64 | int] = None):
        if t is None:
            import time

            t = time.time()
        t = np.int64(t)
        self._pt.attrs["UNIXEND"] = t
        if (
            "UNIXSTR" in self._pt.attrs.keys()
            and "EXPTIME" not in self._pt.attrs.keys()
        ):
            self._pt.attrs["EXPTIME"] = t - self._pt.attrs["UNIXSTR"]
        return self

    def with_exposure_time(self, t: np.int64 | int):
        self._pt.attrs["EXPTIME"] = np.int64(t)
        return self

    def with_flags(self, flags: tuple[str, ...] = DEFAULT_FLAGS):
        import pickle

        pick = np.array(pickle.dumps(flags, 0), dtype="S")
        self._pt.attrs["flags"] = pick
        return self

    def with_energy_resolution(self, resolution: float | np.float64 = 0.1):
        self._pt.attrs["energy_resolution"] = np.float64(resolution)
        return self

    def with_dead_time(self, dead_time: int | np.int64 = 0):
        self._pt.attrs["dead_time"] = np.int64(dead_time)
        return self

    def with_data_dir(self, dir: Optional[str]):
        self._pt.attrs["data_path"] = _pyts(dir)
        return self

    def with_wavecal(self, wavecal: Optional[str] = None):
        self._pt.attrs["wavecal"] = _pyts(wavecal)
        return self

    def with_speccal(self, speccal: Optional[str] = None):
        self._pt.attrs["speccal"] = _pyts(speccal)
        return self

    def with_flatcal(self, flatcal: Optional[str] = None):
        self._pt.attrs["flatcal"] = _pyts(flatcal)
        return self

    def with_cosmiccal(self, cosmiccal: bool = False):
        self._pt.attrs["cosmiccal"] = np.uint8(cosmiccal)
        return self

    def with_lincal(self, lincal: bool = False):
        self._pt.attrs["lincal"] = np.uint8(lincal)
        return self

    def with_pixcal(self, pixcal: bool = False):
        self._pt.attrs["pixcal"] = np.uint8(pixcal)
        return self

    def close(self):
        if self._opened:
            if "UNIXEND" not in self._pt.attrs.keys():
                self.with_end_time()
            self.f.close()
            self._opened = False

    def __del__(self):
        self.close()


def roundtrip(input_file: str, output_file: str):
    import pickle

    fi = h5py.File(input_file)
    pt = fi["photons"]["photontable"]
    pta = dict(pt.attrs)
    fo = (
        TableWriter(output_file, (pta["min_wavelength"], pta["max_wavelength"]))
        .with_data_dir(_styp(pta["data_path"]))
        .with_cosmiccal(bool(pta["cosmiccal"]))
        .with_lincal(bool(pta["lincal"]))
        .with_pixcal(bool(pta["pixcal"]))
        .with_wavecal(_styp(pta["wavecal"]))
        .with_speccal(_styp(pta["speccal"]))
        .with_flatcal(_styp(pta["flatcal"]))
        .with_start_time(pta["UNIXSTR"])
        .with_end_time(pta["UNIXEND"])
        .with_energy_resolution(pta["energy_resolution"])
        .with_dead_time(pta["dead_time"])
        .with_flags(pickle.loads(pta["flags"]))
        .append_photons(pt[::])
    )
    if "beammap" in fi.keys():
        fo.with_beammap(
            _styp(pta["E_BMAP"]), fi["beammap"]["map"], fi["beammap"]["flag"]
        )
    fo.close()



import re
import importlib.resources
from logging import getLogger
from mkidgen3.mkidpynq import get_board_name

_clock_reference = None
_source = None

def _parse_ticspro(file):
    with open(file, 'r') as f:
        lines = [l.rstrip("\n") for l in f]

        registers = []
        for i in lines:
            m = re.search('[\t]*(0x[0-9A-F]*)', i)
            registers.append(int(m.group(1), 16), )
    return registers


def _patch_xrfclk_lmk():
    # access with     xrfdc.set_ref_clks(lmk_freq='122.88_viaext10M')
    import xrfclk

    _LMK04208_FILES = {
        '122.88_viaext10M': 'config/ZCU111_LMK04208_10MHz_Ref_J109SMA.txt'
    }

    _LMK04828_FILES = {
        '256.0_MTS': 'config/LMK04828_256.0_MTS.txt',
        '512.0_MTS': 'config/LMK04828_512.0_MTS.txt',
        '512.0_MTS_dualloop': 'config/LMK04828_512_dualloop.txt',
        '4096.0_MTS': 'config/LMK04828_512_dualloop.txt',
        '500.0_MTS': 'config/LMK04828_500.0_MTS.txt'
    }

    _LMX2594_FILES = {
        '500.0_MTS': 'config/LMX2594_500.0_MTS.txt',
        '512.0_MTS': 'config/LMX2594_512.0_MTS.txt',
        '512.0_MTS_dualloop': 'config/LMX2594_512.0_MTS.txt',
        '4096.0_MTS': 'config/LMX2594_4096.0.txt',
        '409.6_MTS': 'config/LMX2594_409.6_256FoscMTS.txt'
    }

    _CLOCK_CONFIG_DICT = {
    'lmk04208': _LMK04208_FILES,
    'lmk04828': _LMK04828_FILES,
    'lmx2594': _LMX2594_FILES
    }

    for clock_part in _CLOCK_CONFIG_DICT:
        for programming_key, fname in _CLOCK_CONFIG_DICT[clock_part].items():
            tpro_file = importlib.resources.files('mkidgen3').joinpath(fname)
            xrfclk.xrfclk._Config[clock_part][programming_key] = _parse_ticspro(tpro_file)


def configure(programming_key: str | None = None, clock_source: str| None = None):
    """

    Args:
        programming_key: Ignored on the ZCU111.
            '4.096GSPS_MTS' MTS compatible with 4.096 GSPS Sampling Fequency (RFSoC4x2 Only)
            '4.096GSPS_MTS_dualloop'
            '4.096GSPS_MTS_direct'
            '5.000GSPS_MTS' MTS compatible with 5.000 GSPS Sampling Frequency (RFSoC4x2 Only)
        clock_source: internal | external | None. External pulls LMK clock source from 10 MHz Ref

    Returns: Nothing

    """
    global _clock_reference, _source

    programming_key = '' if programming_key is None else programming_key
    clock_source = 'internal' if clock_source is None else clock_source

    try:
        import xrfclk, xrfdc
    except ImportError:
        getLogger(__name__).warning('xrfclk/xrfdc unavaiable, clock will not be started')
        return
    if programming_key:
        _patch_xrfclk_lmk()

    board_name = get_board_name()

    if board_name == 'RFSoC4x2':
        if programming_key == '4.096GSPS_MTS':
            xrfclk.set_ref_clks(lmk_freq='512.0_MTS', lmx_freq='512.0_MTS')
        elif programming_key == '4.096GSPS_MTS_dualloop':
            xrfclk.set_ref_clks(lmk_freq='512.0_MTS_dualloop', lmx_freq='512.0_MTS_dualloop')
        elif programming_key == '4.096GSPS_MTS_direct':
            xrfclk.set_ref_clks(lmk_freq='4096.0_MTS', lmx_freq='4096.0_MTS')
        elif programming_key == '5.000GSPS_MTS':
            xrfclk.set_ref_clks(lmk_freq='500.0_MTS', lmx_freq='500.0_MTS')
        else:
            getLogger(__name__).debug('No programming key specified, defaulting to lmk freq: 245.76 MHz, lmx freq: 409.6 MHz')
            xrfclk.set_ref_clks(lmk_freq=245.76, lmx_freq=409.6)
        for lmk in xrfclk.lmk_devices:
            if clock_source == 'external':
                getLogger(__name__).debug('setting lmk to use external 10 MHz reference')
                xrfclk.xrfclk._write_LMK_regs([0x1470a], lmk)
            if clock_source == 'internal':
                getLogger(__name__).debug('setting lmk to use internal 10 MHz reference')
                xrfclk.xrfclk._write_LMK_regs([0x1471a], lmk)

    elif board_name == 'ZCU111':
        if clock_source == 'external':
            xrfclk.set_ref_clks(lmk_freq='122.88_viaext10M', lmx_freq=409.6)
        else:
            xrfclk.set_ref_clks(lmk_freq=122.88, lmx_freq=409.6)

    else:
        raise ValueError('Unknown board name. Cannot proceed with clock programming.')

    _clock_reference = clock_source
    _source = programming_key


def get_clock():
    return _source, _clock_reference



import pynq


class Quirk:
    """A class to unify working around hardware/pynq quirks in different versions

    Indicate quirk presence with .quirkname properties, provide fixes with .do_quirkname(cls, ..)
    classmethods, and indicate fix completion with .done_quirkname class properties

    For groups of quirks use .pre_eventname() and .post_eventname() classmethods to manage them
    """


class MTS(Quirk):
    """Unclear why this is required but it has been in every bistream I've used, report upstream"""

    double_sync = True


class Overlay(Quirk):

    def __init__(self, overlay:pynq.Overlay):
        self._ol = overlay

    @property
    def interrupt_mangled(self):
        """Work around an inconsistency in name mangling between the HWH parser and the pynq
        interrupt logic

        Report upstream to Xilinx
        """
        for n in self._ol.ip_dict.keys():
            if "axi_intc" in n and "/" in n:
                return True
        return False

    def do_interrupt_mangled(self):
        parser_intcs = self._ol.device.parser.interrupt_controllers
        top_intcs = self._ol.interrupt_controllers
        for n in self._ol.ip_dict.keys():
            if "axi_intc" in n and "/" in n:
                parser_intcs[n] = parser_intcs[n.replace("/", "_")]
                top_intcs[n] = top_intcs[n.replace("/", "_")]

    @property
    def threepart_ddc(self):
        """The three part DDC uses an AXI BRAM controller which is picked up as a memory
        instead of an IP by PYNQ3, this instantiates the 3 part ddc driver so that to the
        user it still looks like an IP
        """
        for n in self._ol.mem_dict.keys():
            if "reschan" in n and "axi_bram_ctrl" in n:
                return True
        return False

    def do_threepart_ddc(self):
        import mkidgen3

        for n in self._ol.mem_dict.keys():
            if "reschan" in n and "axi_bram_ctrl" in n:
                from mkidgen3.drivers.ddc import ThreepartDDC

                num = int(n.split("_")[-1])
                memname = "".join(n.split("/"))
                mmio = getattr(self._ol, memname).mmio
                hier = self._ol
                for h in n.split("/")[:-1]:
                    hier = getattr(hier, h)
                setattr(hier, "ddccontrol_{:d}".format(num), ThreepartDDC(mmio))

    def post_configure(self):
        if not hasattr(self._ol, 'dac_table') and hasattr(self._ol, 'dactable'):
            self._ol.dac_table = self._ol.dactable
        if self.interrupt_mangled:
            self.do_interrupt_mangled()
        if self.threepart_ddc:
            self.do_threepart_ddc()



from pynq import DefaultIP
from logging import getLogger
from mkidgen3.opfb import opfb_bin_number


class BinToResIP(DefaultIP):
    resmap_addr = 0x1000
    bindto = ['MazinLab:mkidgen3:bin_to_res:0.6', 'mazinlab:mkidgen3:bin_to_res:1.33']

    def __init__(self, description):
        """
        The core uses an array of 256 values, each consisting of 8 12 bit numbers packed into 96 bit word that
        specifies the OPFB bin number (0-4095, hence 12bit) used by each of the resonator channels in a group of 8.
        So to drive resonator channel 1035 from OPFB bin 3043 you need to set bits 35:24 (output 3) of group 129 to
        3043.

        control
        0x1000 ~
        0x1fff : Memory 'rid_to_bin' (256 * 128b)
                 Word 4n   : bit [31:0] - rid_to_bin[n][31: 0]
                 Word 4n+1 : bit [31:0] - rid_to_bin[n][63:32]
                 Word 4n+2 : bit [31:0] - rid_to_bin[n][95:64]
                 Word 4n+3 : bit [31:0] - rid_to_bin[n][127:96]  (unused)
        """
        super().__init__(description=description)

    @staticmethod
    def _checkgroup(group_ndx):
        if group_ndx < 0 or group_ndx > 255:
            raise ValueError('group_ndx must be in [0,255]')

    def _read_group(self, group_ndx):
        self._checkgroup(group_ndx)
        g = 0
        vals = [self.read(self.resmap_addr + 16 * group_ndx + 4 * i) for i in range(3)]
        for i, v in enumerate(vals):
            # print(format(v,'032b'))
            g |= v << (32 * i)
        # print('H-'+format(g,'096b')+'-L')
        return [((g >> (12 * j)) & 0xfff) for j in range(8)]

    def _write_group(self, group_ndx, group):
        self._checkgroup(group_ndx)
        if len(group) != 8:
            raise ValueError('len(group)!=8')
        bits = 0
        for i, g in enumerate(group):
            bits |= (int(g) & 0xfff) << (12 * i)
        data = bits.to_bytes(12, 'little', signed=False)
        self.write(self.resmap_addr + 16 * group_ndx, data)

    def bin(self, res):
        """
        Retrieve the OPFB bin assigned to resonator channel res, 0-2047
        The mapping for resonator i is 12 bits and will require reading 1 or 2 32 bit word
        n=i//8 j=(i%8)*12//32
        """
        return self._read_group(res // 8)[res % 8]

    @property
    def bins(self):
        """ Return a tuple of the OPFB bins assigned to each of the 2048 resonator channels"""
        return tuple([v for g in range(256) for v in self._read_group(g)])

    @bins.setter
    def bins(self, bins):
        """ Set which OPFB bins to use for each resonator channel. Set to list of 2048 numbers in [0, 4095]"""
        if len(bins) != 2048:
            raise ValueError('len(bins)!=2048')
        if min(bins) < 0 or max(bins) > 4095:
            raise ValueError('Bin values must be in [0,4095]')
        for i in range(256):
            self._write_group(i, bins[i * 8:i * 8 + 8])

    def configure(self, bins=None):
        getLogger(__name__).info(f'Configuring bin2res with bins {bins}')
        self.bins = bins
